#!/usr/bin/env python3
"""

"""
# ruff: noqa: B011
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

from jgdv.structs.locator.errors import DirAbsent, LocationExpansionError, LocationError
from jgdv.structs.locator import JGDVLocator, Location
from jgdv.structs.locator.locator import _LocatorGlobal
from jgdv.structs.dkey import DKey, NonDKey

logging = logmod.root

match JGDVLocator.Current:
    case None:
        initial_loc = JGDVLocator(pl.Path.cwd())
    case x:
        initial_loc = x

@pytest.fixture(scope="function")
def simple() -> JGDVLocator:
    return JGDVLocator(pl.Path.cwd())

class TestLocator_Get:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_get_registered(self, simple):
        simple.update({"blah":"bloo/blee"})
        match simple.get("blah"):
            case pl.Path() as x if x == pl.Path("bloo/blee"):
                assert(True)
            case x:
                assert(False), x

    def test_get_missing(self, simple):
        simple.update({"blah":"bloo/blee"})
        with pytest.raises(KeyError):
            simple.get("aweg")

    def test_get_fallback(self, simple):
        simple.update({"blah":"bloo/blee"})
        match simple.get("aweg", "bloo"):
            case pl.Path() as x if x == pl.Path("bloo"):
                assert(True)
            case x:
                assert(False), x

class TestLocator_Access:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_registered(self, simple):
        simple.update({"blah":"bloo/blee"})
        match simple.access("blah"):
            case Location():
                assert(True)
            case x:
                assert(False), x

    def test_missing(self, simple):
        simple.update({"blah":"bloo/blee"})
        match simple.access("aweg"):
            case None:
                assert(True)
            case x:
                assert(False), x

class TestLocator_Expand:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_nokey(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "aweg"
        match simple.expand("aweg", strict=False):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_single_registered(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_single_missing_strict(self, simple):
        simple.update({"blah":"bloo/blee"})
        with pytest.raises(KeyError):
            simple.expand("{aweg}", strict=True)

    def test_single_missing_not_strict(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "{aweg}"
        match simple.expand("{aweg}", strict=False):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_single_missing_fallback(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_multi_registered(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_multi_missing_strict(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_multi_missing_not_strict(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_multi_missing_fallback(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_simple_recursion(self, simple):
        simple.update({"blah":"bloo/blee"})
        target = pl.Path.cwd() / "bloo/blee"
        match simple.expand("{blah}"):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

    def test_double_recursion(self, simple):
        simple.update({"blah":"{bloo}/blee", "bloo":"aweg"})
        target = pl.Path.cwd() / "aweg/blee"
        match simple.expand("{blah}", strict=False):
            case pl.Path() as x if x == target:
                assert(True)
            case x:
                assert(False), x

class TestLocator_MainAccess:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_attr_basic_dir(self, simple):
        """Locator.blah            -> {cwd}/ex/dir"""
        target = Location("ex/dir")
        simple.update({"blah": "ex/dir"})
        assert("blah" in simple)
        assert(simple.blah == target)

    def test_item_basic_dir(self, simple):
        """Locator['{blah}']       -> {cwd}/ex/dir"""
        target = pl.Path.cwd() / "ex/dir"
        simple.update({"blah": "ex/dir"})
        assert("blah" in simple)
        assert(simple['{blah}'] == target)

    def test_item_join_dir(self, simple):
        """Locator['{blah}/blee']  -> {cwd}/ex/dir/blee"""
        target = pl.Path.cwd() / "ex/dir/blee"
        simple.update({"blah": "ex/dir"})
        assert("blah" in simple)
        assert(simple['{blah}/blee'] == target)

    def test_attr_basic_file(self, simple):
        """Locator.bloo            -> {cwd}/a/b/c.txt"""
        target = Location("file::>a/b/c.txt")
        simple.update({"bloo": "file::>a/b/c.txt"})
        assert("bloo" in simple)
        assert(simple.bloo == target)

    def test_item_basic_file(self, simple):
        """Locator['{bloo}']       -> {cwd}/a/b/c.txt"""
        target = pl.Path.cwd() / "a/b/c.txt"
        simple.update({"bloo": "file::>a/b/c.txt"})
        assert("bloo" in simple)
        assert(simple['{bloo}'] == target)

    def test_item_dir_file_join(self, simple):
        """ Locator[a/b/c] -> {cwd}/a/b/c """
        target = pl.Path.cwd() / "a/b/c"
        simple.update({"bloo": "dir::>a/b/c"})
        assert("bloo" in simple)
        assert(simple['{bloo}'] == target)

    def test_item_nonkey(self, simple):
        """ Locator[bloo/blah] -> {cwd}/bloo/blah """
        target = pl.Path.cwd() / "bloo/blah"
        simple.update({"bloo": "dir::>a/b/c"})
        assert("bloo" in simple)
        assert(simple['bloo/blah'] == target)

class TestLocation_ExpansionConflict:

    @pytest.mark.xfail
    def test_item_file_join_fail(self, simple):
        """Locator['{bloo}/blee']  -> Error"""
        simple.update({"bloo": "file::>a/b/c.txt"})
        assert("bloo" in simple)
        with pytest.raises(LocationError):
            simple['{bloo}/blah']



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast, Self, Final)
import warnings

import pytest

logging = logmod.root

from ... import DKey, Key_p
from ..args_keys import ArgsDKey, KwargsDKey

IMP_KEY_BASES               : Final[list[str]]           = ["bob", "bill", "blah", "other", "23boo", "aweg2531", "awe_weg", "aweg-weji-joi"]
EXP_KEY_BASES               : Final[list[str]]           = [f"{{{x}}}" for x in IMP_KEY_BASES]
EXP_P_KEY_BASES             : Final[list[str]]           = ["{bob:wd}", "{bill:w}", "{blah:wi}", "{other:i}"]
PATH_KEYS                   : Final[list[str]]           = ["{bob}/{bill}", "{blah}/{bloo}", "{blah}/{bloo}"]
MUTI_KEYS                   : Final[list[str]]           = ["{bob}_{bill}", "{blah} <> {bloo}", "! {blah}! {bloo}!"]
IMP_IND_KEYS                : Final[list[str]]           = ["bob_", "bill_", "blah_", "other_"]
EXP_IND_KEYS                : Final[list[str]]           = [f"{{{x}}}" for x in IMP_IND_KEYS]

VALID_KEYS                                           = IMP_KEY_BASES + EXP_KEY_BASES + EXP_P_KEY_BASES + IMP_IND_KEYS + EXP_IND_KEYS
VALID_MULTI_KEYS                                     = PATH_KEYS + MUTI_KEYS

class TestArgsDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_mark(self):
        assert(DKey.MarkOf(ArgsDKey) is DKey.Marks.ARGS)

    @pytest.mark.skip
    def test_todo(self):
        pass

class TestKwargsDKey:

    def test_mark(self):
        assert(DKey.MarkOf(KwargsDKey) is DKey.Marks.KWARGS)

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""



"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

from jgdv import Maybe, Lambda
from jgdv.decorators import Decorator

from ._interface import Logger, LOGDEC_PRE

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Global Vars:

# Body:

class LogCall(Decorator):
    """ A Decorator for announcing the entry/exit of a function call

    eg:
    @LogCall(enter="Entering", exit="Exiting", level=logmod.INFO)
    def a_func()...
    """

    def __init__(self, enter:Maybe[str|Lambda]=None, exit:Maybe[str|Lambda]=None, level:int|str=logmod.INFO, logger:Maybe[Logger]=None):
        super().__init__(prefix=LOGDEC_PRE)
        self._logger = logger or logging
        self._enter_msg = enter
        self._exit_msg = exit
        match level:
            case str():
                self._level = logmod.getLevelNamesMapping().get(level, logmod.INFO)
            case int():
                self._level = level
            case _:
                raise ValueError(level)

    def _log_msg(self, msg:Maybe[str|Lambda], fn:Callable, args:list, **kwargs):
        match msg:
            case None:
                return None
            case types.FunctionType():
                msg = msg(fn, *args, **kwargs)
            case str():
                pass
            case _:
                raise TypeError(msg)

        self._logger.log(self.level)

    def _wrap_method(self, fn) -> Callable:

        def basic_wrapper(_self, *args, **kwargs):
            self._log_msg(self._enter_msg, fn, args, obj=_self, **kwargs)
            ret_val = fn(_self, *args, **kwargs)
            self._log_msg(self._exit_msg, fn, args, obj=_self, returned=ret_val, **kwargs)
            return ret_val

        return basic_wrapper

    def _wrap_fn(self, fn) -> Callable:

        def basic_wrapper(*args, **kwargs):
            self._log_msg(self._enter_msg, fn, args, obj=None, **kwargs)
            ret_val = fn(*args, **kwargs)
            self._log_msg(self._exit_msg, fn, args, obj=None, returned=ret_val, **kwargs)
            return ret_val

    def _wrap_class(self, cls) -> type:
        raise NotImplementedError()



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import sys
import time
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from types import TracebackType

if TYPE_CHECKING:
    from jgdv import Maybe, Traceback, Frame
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

if not hasattr(sys, "_getframe"):
        msg = "Can't use TraceBuilder on this system, there is no sys._getframe"
        raise ImportError(msg)

##--|
class TraceBuilder:
    """ A Helper to simplify access to tracebacks.
    By Default, removes the frames of this tracebuilder from the trace
    ie     : TraceBuilder._get_frames() -> TraceBuilder.__init__() -> call -> call -> root
    will be: call -> call -> root

    use item acccess to limit the frames,
    eg: tb[2:], will remove the two most recent frames from the traceback

    Use as:
    tb = TraceBuilder()
    raise Exception().with_traceback(tb[:])
    """


    def __class_getitem__(cls, item:slice) -> Traceback:
        tbb = cls()
        return tbb[item]

    def __init__(self, *, chop_self:bool=True) -> None:
        self.frames : list[Frame] = []
        self._get_frames()
        if chop_self:
            self.frames = self.frames[2:]

    def __getitem__(self, val:Maybe[slice]=None) -> Traceback:
        match val:
            case None:
                return self.to_tb()
            case slice() | int():
                return self.to_tb(self.frames[val])
            case _:
                msg = "Bad value passed to TraceHelper"
                raise TypeError(msg, val)

    def _get_frames(self) -> None:
        """ from https://stackoverflow.com/questions/27138440
        Builds the frame stack from most recent to least,
        """
        depth = 0
        while True:
            try:
                frame : Frame = sys._getframe(depth)
                depth += 1
            except ValueError:
                break
            else:
                self.frames.append(frame)

    def to_tb(self, frames:Maybe[list[Frame]]=None) -> Traceback:
        top    = None
        frames = frames or self.frames
        for frame in frames:
            top = TracebackType(top, frame,
                                frame.f_lasti,
                                frame.f_lineno)
        else:
            return top



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import string
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv._abstract.protocols import Buildable_p, SpecStruct_p
from jgdv.structs.strang import CodeReference

# ##-- end 1st party imports

from .._interface import DKeyMark_e
from .. import DKey
from ..keys import MultiDKey, NonDKey, SingleDKey

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Ident, RxStr, Rx
    from .._interface import KeyMark

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class StrDKey(DKey, mark=str, convert="s"):
    """
    A Simple key that always expands to a string
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data.expansion_type  = str
        self.data.typecheck       = str



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from typing import Any
from uuid import UUID, uuid1

# ##-- end stdlib imports

import jgdv
from jgdv.decorators import MetaDec

##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Traceback
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger = logmod.Logger
##--|

# isort: on
##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class TimeBlock_ctx(jgdv.protos.DILogger_p):
    """
    A Simple Timer Ctx class to log how long things take
    Give it a logger, a message, and a level.
    The message doesn't do any interpolation

    eg: With TimeBlock():...

    """
    start_time   : Maybe[float]
    end_time     : Maybe[float]
    elapsed_time : Maybe[float]

    def __init__(self, *, logger:Maybe[Logger|False]=None, enter:Maybe[str]=None, exit:Maybe[str]=None, level:Maybe[int|str]=None) -> None:
        self.start_time      = None
        self.end_time        = None
        self.elapsed_time    = None
        self._enter_msg      = enter or "Starting Timer"
        self._exit_msg       = exit or "Time Elapsed"

        match level:
            case int() as x:
                self._level = x
            case str() as x if x in logmod._nameToLevel:
                self._level = logmod._nameToLevel[x]
            case _:
                self._level = logmod.INFO

        match logger:
            case False:
                self._logger = logmod.getLogger("null")
                self._logger.propagate = False
                pass
            case logmod.Logger() as l:
                self._logger = l
            case _:
                self._logger = logging


    def __enter__(self) -> Self:
        self._logger.log(self._level, self._enter_msg)
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type:Maybe[type], exc_value:Maybe, exc_traceback:Maybe[Traceback]) -> bool:
        self.end_time = time.perf_counter()
        self.elapsed_time = self.end_time - self.start_time
        self._logger.log(self._level, "%s : %s", self._exit_msg, f"{self.elapsed_time:0.4f} Seconds")
        # return False to reraise errors
        return False


class TrackTime(MetaDec):
    """ Decorate a callable to track its timing """

    def __init__(self, logger:Maybe[Logger]=None, level:Maybe[int|str]=None, enter:Maybe[str]=None, exit:Maybe[str]=None, **kwargs:Any) -> None:  # noqa: A002, ANN401
        kwargs.setdefault("mark", "_timetrack_mark")
        kwargs.setdefault("data", "_timetrack_data")
        super().__init__([], **kwargs)
        self._logger = logger
        self._level  =  level
        self._entry  = entry
        self._exit   = exit

    def wrap_fn[I, O](self, fn:Func[I, O]) -> Func[I, O]:
        logger, enter, exit, level = self._logger, self._entry, self.exit, self.level  # noqa: A001

        def track_time_wrapper(*args:Any, **kwargs:Any) -> O:  # noqa: ANN401
            with TimeBlock_ctx(logger=logger, enter=enter, exit=exit, level=level):
                return fn(*args, **kwargs)

        return track_time_wrapper

    def wrap_method[I, O](self, fn:Method[I, O]) -> Method[I, O]:
        return self._wrap_fn(fn)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ARG002
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import atexit# for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from collections import defaultdict, deque
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
# ##-- end stdlib imports

# ##-- 3rd party imports
import sh

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import identity_fn, Proto
from jgdv.decorators import DoMaybe
from jgdv.structs.strang import CodeReference, Strang
from .. import _interface as API # noqa: N812
# ##-- end 1st party imports

from . import _interface as ExpAPI # noqa: N812
from ._interface import Expander_p, ExpInst_d, SourceChain_d

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, Self, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, overload
from collections.abc import Callable

if TYPE_CHECKING:
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, M_, Func, RxStr, Rx, Ident, FmtStr, Ctor
    from ._interface import Expandable_p
    from .._interface import Key_p
##--|
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
type ExpOpts          = ExpAPI.ExpOpts
type InstructionAlts  = ExpAPI.InstructionAlts
type InstructionList  = ExpAPI.InstructionList

# Body:

@Proto(Expander_p[API.Key_p])
class DKeyExpander:
    """ A Static class to control expansion.

    In order it does::

        - pre-format the value to (A, coerceA,B, coerceB)
        - (lookup A) or (lookup B) or None
        - manipulates the retrieved value
        - potentially recurses on retrieved values
        - type coerces the value
        - runs a post-coercion hook
        - checks the type of the value to be returned

    During the above, the hooks of Expandable_p will be called on the source,
    if they return nothing, the default hook implementation is used.

    All of those steps are fallible.
    When one of them fails, then the expansion tries to return, in order::

        - a fallback value passed into the expansion call
        - a fallback value stored on construction of the key
        - None

    Redirection Rules::

        - Hit          || {test}  => state[test=>blah]  => blah
        - Soft Miss    || {test}  => state[test_=>blah] => {blah}
        - Hard Miss    || {test}  => state[...]         => fallback or None

    Indirect Keys act as::

        - Indirect Soft Hit ||  {test_}  => state[test_=>blah] => {blah}
        - Indirect Hard Hit ||  {test_}  => state[test=>blah]  => blah
        - Indirect Miss     ||  {test_} => state[...]          => {test_}

    """

    _ctor : Ctor[API.Key_p]

    def __init__(self) -> None:
        self._ctor = None # type: ignore[assignment]

    def set_ctor(self, ctor:Ctor[API.Key_p]) -> None:
        """ Dependency injection from DKey.__init_subclass__ """
        if self._ctor is not None:
            return

        self._ctor = ctor

    ##--|

    def redirect(self, source:API.Key_p, *sources:ExpAPI.SourceBases, **kwargs:Any) -> list[Maybe[ExpInst_d]]:  # noqa: ANN401
            return [self.expand(source, *sources, limit=1, **kwargs)]

    def expand(self, key:API.Key_p, *sources:ExpAPI.SourceBases|SourceChain_d, **kwargs:Any) -> Maybe[ExpInst_d]:  # noqa: ANN401, PLR0912
        """ The entry point for expanding a key """
        full_sources  : SourceChain_d
        targets       : InstructionAlts
        alts          : Maybe[InstructionList]
        value         : Maybe[ExpInst_d]
        fallback      : Maybe[ExpInst_d]

        ##--|
        logging.info("- Expanding: [%s]", repr(key))
        if key.MarkOf(key) is False:
            return ExpInst_d(value=key, literal=True)

        match kwargs.get("fallback", key.data.fallback):
            case None:
                fallback = None
            case type() as ctor:
                x = ctor()
                fallback = ExpInst_d(value=x, literal=True)
            case ExpInst_d() as x:
                fallback = x
            case x:
                fallback = ExpInst_d(value=x, literal=True)

        # Limit defaults to -1 / until completion
        # but recursions can pass in limits
        match kwargs.get("limit", key.data.max_expansions):
            case ExpAPI.NO_EXPANSIONS_PERMITTED:
                return fallback or ExpInst_d(value=key, literal=True)
            case None | ExpAPI.UNRESTRICTED_EXPANSION:
                limit = ExpAPI.UNRESTRICTED_EXPANSION
            case x if x < ExpAPI.UNRESTRICTED_EXPANSION:
                limit = ExpAPI.UNRESTRICTED_EXPANSION
            case x:
                # decrement the limit
                limit  = x - 1

        full_sources   = self.extra_sources(sources, source=key)
        targets        = self.generate_alternatives(full_sources, kwargs, source=key)
        # These are Maybe monads:
        alts       = self.do_lookup(targets, full_sources, kwargs, source=key)
        alts       = self.configure_recursion(alts, full_sources, kwargs, source=key)
        alts       = self.do_recursion(alts, full_sources, kwargs, max_rec=limit, source=key)
        value      = self.flatten(alts, kwargs, source=key)
        value      = self.coerce_result(value, kwargs, source=key)
        self.check_result(key, value, kwargs)
        match value:
            case None if fallback:
                logging.debug("|-| %s -> None, Fallback: %s", key, fallback)
                return self.finalise(fallback, kwargs, source=key)
            case None:
                return None
            case ExpInst_d(literal=False) as x:
                msg = "Expansion didn't result in a literal"
                raise ValueError(msg, x, key)
            case ExpInst_d() as x:
                logging.info("|-| %s -> %s", key, x)
                return self.finalise(x, kwargs, source=key)
            case x:
                raise TypeError(type(x))

    ##--| Expansion phases

    def extra_sources(self, sources:Iterable[ExpAPI.SourceBases|SourceChain_d], *, source:API.Key_p) -> SourceChain_d:
        x : Any
        mapping : SourceChain_d
        match sources:
            case [SourceChain_d() as x]:
                mapping = x
            case [*xs]:
                mapping = SourceChain_d(*xs, lifter=self._ctor)

        if not hasattr(source, "exp_extra_sources_h"):
            return mapping

        match source.exp_extra_sources_h(mapping):
            case SourceChain_d() as x:
                return x
            case x:
                raise TypeError(type(x))

    def generate_alternatives(self, sources:SourceChain_d, opts:ExpOpts, *, source:API.Key_p) -> InstructionAlts:
        """
        returns a list (L1) of lists (L2) of target tuples (T).
        When looked up, For each L2, the first T that returns a value is added
        to the final result
        """
        if not hasattr(source, "exp_generate_alternatives_h"):
            return [[
                source.to_exp_inst(),
                source.to_exp_inst(indirect=True, lift=True),
            ]]

        match source.exp_generate_alternatives_h(sources, opts):
            case [] | None:
                return [[
                    source.to_exp_inst(),
                    source.to_exp_inst(indirect=True, lift=True),
                ]]
            case list() as xs:
                return xs
            case x:
                raise TypeError(type(x))

    @DoMaybe()
    def do_lookup(self, targets:InstructionAlts, sources:SourceChain_d, opts:ExpOpts, *, source:API.Key_p) -> Maybe[InstructionList]:
        """ customisable method for each key subtype
            Target is a list (L1) of lists (L2) of target tuples (T).
            For each L2, the first T that returns a value is added to the final result
            """
        target : list[ExpInst_d]
        result = []
        logging.debug("- Lookup: %s", targets)
        for target in targets:
            match sources.lookup(target):
                case None:
                    logging.debug("Lookup Failed for: %s", target)
                    return []
                case ExpInst_d(value=API.Key_p() as key, rec=ExpAPI.UNRESTRICTED_EXPANSION) as res if source == key:
                    # guard against infinit recursion
                    res.rec = API.RECURSION_GUARD
                    result.append(res)
                case ExpInst_d() as x:
                    result.append(x)
                case x:
                    msg = "LookupTarget didn't return an ExpInst_d"
                    raise TypeError(msg, x)
        else:
            return result

    @DoMaybe()
    def configure_recursion(self, insts:InstructionList, sources:SourceChain_d, opts:ExpOpts, *, source:API.Key_p) -> Maybe[InstructionList]:
        """ Produces a list[Key|Val|(Key, rec:int)]"""
        if not hasattr(source, "exp_configure_recursion_h"):
            return insts

        match source.exp_configure_recursion_h(insts, sources, opts):
            case None:
                return insts
            case list() as newinsts:
                return newinsts
            case x:
                raise TypeError(type(x))

    @DoMaybe()
    def do_recursion(self, insts:InstructionList, sources:SourceChain_d, opts:ExpOpts, *, max_rec:int=API.RECURSION_GUARD, source:API.Key_p) -> Maybe[InstructionList]:  # noqa: PLR0912
        """
        For values that can expand futher, try to expand them

        """
        x : Any
        recurse_on  : Maybe[API.Key_p]
        result      : list[ExpInst_d]  = []

        if not bool(insts):
            return result
        logging.debug("- Recursing: %r : %s : %s", source, insts, max_rec)
        for inst in insts:
            recurse_on  = None
            # Decide if there should be a recursion
            match inst:
                case ExpInst_d(literal=True) | ExpInst_d(rec=ExpAPI.NO_EXPANSIONS_PERMITTED) as res:
                    result.append(res)
                case ExpInst_d(value=API.Key_p() as key, rec=ExpAPI.UNRESTRICTED_EXPANSION) if key is source or key == source:
                    msg = "Unrestrained Recursive Expansion"
                    raise RecursionError(msg, source)
                case ExpInst_d(value=API.Key_p() as key, rec=ExpAPI.UNRESTRICTED_EXPANSION):
                    recurse_on = key
                case ExpInst_d(value=API.Key_p() as key):
                    recurse_on  = key
                case ExpInst_d(value=pl.Path()|str() as key) if bool:
                    recurse_on = self._ctor(key)
                # case ExpInst_d(value=key):
                #     recurse_on  = self._ctor(key)
                case ExpInst_d() as res: # Protect against nonsensical coercions
                    result.append(res)
                case x:
                    msg = "Unexpected Recursion Value"
                    raise TypeError(msg, x)

            # do the recursion
            if recurse_on is None:
                continue

            match inst.rec, max_rec:
                case ExpAPI.UNRESTRICTED_EXPANSION, x:
                    rec_limit = x
                case x, ExpAPI.UNRESTRICTED_EXPANSION:
                    rec_limit = x
                case int() as x, int() as y:
                    rec_limit = min(x, y)
            logging.debug("----")
            match self.expand(recurse_on, sources, limit=rec_limit, fallback=inst.fallback, convert=inst.convert):
                case None if inst.lift:
                    result.append(ExpInst_d(value=recurse_on, literal=True))
                case None:
                    pass
                case ExpInst_d() as exp if inst.lift:
                    exp.convert = False
                    result.append(exp)
                case ExpInst_d() as exp:
                    result.append(exp)
                case other:
                    raise TypeError(type(other))
            logging.debug("----")
        else:
            logging.debug("- Recursion Finished: %r : %r", source, result)
            return result

    @DoMaybe()
    def flatten(self, insts:InstructionList, opts:ExpOpts, *, source:API.Key_p) -> Maybe[ExpInst_d]:
        """
        Flatten separate expansions into a single value
        """
        match insts:
            case []:
                return None
            case [x, *_] if not hasattr(source ,"exp_flatten_h"):
                return x
            case _ if not hasattr(source ,"exp_flatten_h"):
                return None
            case _:
                pass

        match source.exp_flatten_h(insts, opts):
            case None:
                return None
            case ExpInst_d() as x:
                return x
            case x:
                raise TypeError(type(x))

    @DoMaybe()
    def coerce_result(self, inst:ExpInst_d, opts:ExpOpts, *, source:API.Key_p) -> Maybe[ExpInst_d]:
        """
        Coerce the expanded value accoring to source's expansion type ctor
        """

        result : Maybe[ExpInst_d] = None

        if hasattr(source, "exp_coerce_h"):
            match source.exp_coerce_h(inst, opts):
                case ExpInst_d() as x:
                    return x
                case None:
                    pass

        ##--|
        match inst:
            case ExpInst_d(convert=False):
                # Conversion is off
                result = inst
            case ExpInst_d(value=value, convert=None) if isinstance(source.data.expansion_type, type) and isinstance(value, source.data.expansion_type):
                # Type is already correct
                result = inst
            case ExpInst_d(value=value, convert=None) if source.data.expansion_type is not identity_fn:
                # coerce a real ctor
                if not isinstance(value, source.data.expansion_type):
                    inst.value = source.data.expansion_type(value)
                result = inst
            case ExpInst_d(convert=None) if source.data.convert is None:
                # No conv params
                result = inst
            case ExpInst_d(convert=str() as conv):
                # Conv params in expinst
                result = (self._coerce_result_by_conv_param(inst, conv, opts, source=source)
                          or inst)
            case ExpInst_d() if source.data.convert:
                #  Conv params in source
                result = (self._coerce_result_by_conv_param(inst, source.data.convert, opts, source=source)
                          or inst)
            case ExpInst_d():
                result = inst
            case x:
                raise TypeError(type(x))

        ##--|
        logging.debug("- Type Coerced: (%r) %s -> %s", source, source.data.convert, result)
        result.literal = True
        return result

    @DoMaybe()
    def finalise(self, inst:ExpInst_d, opts:ExpOpts, *, source:API.Key_p) -> Maybe[ExpInst_d]:
        """
        A place for any remaining modifications of the result or fallback value
        """
        if not hasattr(source, "exp_final_h"):
            inst.literal = True
            return inst

        match source.exp_final_h(inst, opts):
            case None:
                return None
            case ExpInst_d() as x:
                return x
            case x:
                raise TypeError(type(x))

    @DoMaybe()
    def check_result(self, source:API.Key_p, inst:ExpInst_d, opts:ExpOpts) -> None:
        """ check the type of the expansion is correct,
        throw a type error otherwise
        """
        if not hasattr(source, "exp_check_result_h"):
            return

        source.exp_check_result_h(inst, opts)

    ##--| Utils
    def _coerce_result_by_conv_param(self, inst:ExpInst_d, conv:str, opts:ExpOpts, *, source:API.Key_p) -> Maybe[ExpInst_d]:
        """ really, keys with conv params should been built as a
        specialized registered type, to use an exp_final_hook
        """
        match ExpAPI.EXPANSION_CONVERT_MAPPING.get(conv, None):
            case fn if callable(fn):
                val : Any = fn(inst.value)
                return ExpInst_d(value=val)
            case None:
                return inst
            case x:
                logging.warning("Unknown Conversion Parameter: %s", x)
                return None



#!/usr/bin/env python3
"""

"""
# mypy: disable-error-code="attr-defined"
# ruff: noqa: ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import keyword
import logging as logmod
import pathlib as pl
import re
import time
import types as types_
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Mixin
from jgdv.decorators import (
    DataDec,
    MetaDec,
    DecoratorAccessor_m,
    DForm_e,
)
from jgdv.structs.strang import CodeReference
from .. import errors as dkey_errs
from .. import DKey
from .._interface import ARGS_K, KWARGS_K, PARAM_IGNORES, Key_p

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from types import MethodType
from collections.abc import Mapping
from typing import Any

if TYPE_CHECKING:
    from jgdv import Method
    import inspect
    from jgdv import Decorator, FmtStr, Func, Ident, Maybe, Rx
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

    from jgdv.decorators._interface import Decorated, Decorable
    type Signature = inspect.Signature
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DKeyMetaDecorator(MetaDec):
    """ A Meta decorator that registers keys for input and output
    verification"""

    def __init__(self, *args, **kwargs) -> None:
        kwargs.setdefault("mark", "_dkey_meta_marked")
        kwargs.setdefault("data", "_dkey_meta_vals")
        super().__init__(*args, **kwargs)

class DKeyExpansionDecorator(DataDec):
    """
    Utility class for idempotently decorating actions with auto-expanded keys

    """
    _param_ignores : tuple[str, ...]

    def __init__(self, keys:list[DKey], ignores:Maybe[list[str]]=None, **kwargs) -> None:
        kwargs.setdefault("mark", "_dkey_marked")
        kwargs.setdefault("data", "_dkey_vals")
        super().__init__(keys, **kwargs) # type: ignore
        match ignores:
            case None:
                self._param_ignores = PARAM_IGNORES
            case list():
                self._param_ignores = tuple(ignores)
            case x:
                raise TypeError(type(x))

    def _wrap_method_h[**In, Out](self, meth:Method[In,Out]) -> Decorated[In, Out]:
        data_key = self._data_key

        def _method_action_expansions(*call_args:In.args, **kwargs:In.kwargs) -> Out:
            _self, spec, state, *rest = call_args
            try:
                expansions = [x(spec, state) for x in meth.__annotations__[data_key]]
            except KeyError as err:
                logging.warning("Action State Expansion Failure: %s", err)
                return cast("Out", False)  # noqa: FBT003
            else:
                return meth(*call_args, *expansions, **kwargs)

        # -
        return cast("Method", _method_action_expansions)

    def _wrap_fn_h[**In, Out](self, fn:Func[In, Out]) -> Decorated[In, Out]:
        data_key = self._data_key

        def _fn_action_expansions(*args:In.args, **kwargs:In.kwargs) -> Out:
            spec, state, *rest = args
            try:
                expansions = [x(spec, state) for x in fn.__annotations__[data_key]]
            except KeyError as err:
                logging.warning("Action State Expansion Failure: %s", err)
                return cast("Out", False)  # noqa: FBT003
            else:
                return fn(*args, *expansions, **kwargs)

        # -
        return _fn_action_expansions

    def _validate_sig_h(self, sig:Signature, form:DForm_e, args:Maybe[list[DKey]]=None) -> None:
        # Get the head args
        match form:
            case DForm_e.FUNC:
                head = ["spec", "state"]
            case DForm_e.METHOD:
                head = ["self", "spec", "state"]

        params      = list(sig.parameters)
        tail        = args or []

        # Check the head
        for x,y in zip(params, head, strict=False):
            if x != y:
                msg = "Mismatch in signature head"
                raise dkey_errs.DecorationMismatch(msg, x, y, form)

        prefix_ig, suffix_ig = self._param_ignores
        # Then the tail, backwards, because the decorators are applied in reverse order
        for x,y in zip(params[::-1], tail[::-1], strict=False):
            assert(isinstance(y, Key_p))
            key_str = y.var_name()
            if x.startswith(prefix_ig) or x.endswith(suffix_ig):
                logging.debug("Skipping: %s", x)
                continue

            if keyword.iskeyword(key_str):
                msg = "Key is a keyword, use an alias like _{} or {}_ex, or use named={}"
                raise dkey_errs.DecorationMismatch(msg, x, key_str)

            if not key_str.isidentifier():
                msg = "Key is not an identifier, use an alias _{} or {}_ex or use named={}"
                raise dkey_errs.DecorationMismatch(msg, x, key_str)

            if x != y:
                msg = "Mismatch in signature tail"
                raise dkey_errs.DecorationMismatch(msg, str(x), key_str)

class DKeyed:
    """ Decorators for actions

    It registers arguments on an action and extracts them from the spec and state automatically.

    provides: expands/paths/types/requires/returns/args/kwargs/redirects
    The kwarg 'hint' takes a dict and passes the contents to the relevant expansion method as kwargs

    arguments are added to the tail of the action args, in order of the decorators.
    the name of the expansion is expected to be the name of the action parameter,
    with a "_" prepended if the name would conflict with a keyword., or with "_ex" as a suffix
    eg: @DKeyed.paths("from") -> def __call__(self, spec, state, _from):...
    or: @DKeyed.paths("from") -> def __call__(self, spec, state, from_ex):...
    """

    _extensions         : ClassVar[set[type]] = set()
    _decoration_builder : ClassVar[type]      = DKeyExpansionDecorator

    def __init_subclass__(cls) -> None:
        """
        Subclasses of DKeyed are stored, and used to extend DKeyed
        """
        super().__init_subclass__()
        if cls in DKeyed._extensions:
            return
        DKeyed._extensions.add(cls)
        for x in dir(cls):
            match getattr(cls, x):
                case MethodType() as y if not hasattr(DKeyed, x):
                    setattr(DKeyed, x, y)
                case _:
                    pass

class DKeyedMeta(DKeyed):
    """ Subclass extension for decorators that declare meta information,
    but doesnt modify the behaviour
    """

    @classmethod
    def requires(cls, *args, **kwargs) -> DKeyMetaDecorator:
        """ mark an action as requiring certain keys to in the state, but aren't expanded """
        keys = [DKey[Any](x, implicit=True, **kwargs) for x in args]
        return DKeyMetaDecorator(keys)

    @classmethod
    def returns(cls, *args, **kwargs) -> DKeyMetaDecorator:
        """ mark an action as needing to return certain keys """
        keys = [DKey[Any](x, implicit=True, **kwargs) for x in args]
        return DKeyMetaDecorator(keys)

class DKeyedRetrieval(DecoratorAccessor_m, DKeyed):
    """ Subclass extension for DKeyed decorators,
    which modify the calling behaviour of the decoration target

    """

    _decoration_builder : ClassVar[type] = DKeyExpansionDecorator

    @classmethod
    def formats(cls, *args, **kwargs) -> Decorator:
        keys     = [DKey[str](x, implicit=True, **kwargs) for x in args]
        return cls._build_decorator(keys)

    @classmethod
    def expands(cls, *args, **kwargs) -> Decorator:
        """ mark an action as using expanded string keys """
        return cls.formats(*args, **kwargs)

    @classmethod
    def paths(cls, *args, **kwargs) -> Decorator:
        """ mark an action as using expanded path keys """
        kwargs.setdefault("implicit", True)
        keys = [DKey[pl.Path](x, **kwargs) for x in args]
        return cls._build_decorator(keys)

    @classmethod
    def types(cls, *args, **kwargs) -> Decorator:
        """ mark an action as using raw type keys """
        keys : list = [DKey[Any](x, implicit=True, **kwargs) for x in args]
        return cls._build_decorator(keys)

    @classmethod
    def toggles(cls, *args, **kwargs) -> Decorator:
        keys : list = [DKey(x, implicit=True, ctor=bool, check=bool, **kwargs) for x in args]
        return cls._build_decorator(keys)

    @classmethod
    def args(cls, fn:Callable) -> Decorator:
        """ mark an action as using spec.args """
        keys = [DKey[DKey.Marks.ARGS](ARGS_K, implicit=True)] # type: ignore[name-defined]
        match cls._build_decorator(keys)(fn):
            case None:
                raise dkey_errs.DecorationMismatch()
            case x:
                return x

    @classmethod
    def kwargs(cls, fn:Callable) -> Decorator:
        """ mark an action as using all kwargs"""
        keys = [DKey[DKey.Marks.KWARGS](KWARGS_K, implicit=True)] # type: ignore[name-defined]
        match cls._build_decorator(keys)(fn):
            case None:
                raise dkey_errs.DecorationMismatch()
            case x:
                return x

    @classmethod
    def redirects(cls, *args, **kwargs) -> Decorator:
        """ mark an action as using redirection keys """
        kwargs.setdefault("max_exp", 1)
        keys = [DKey[Mapping](x, implicit=True, **kwargs) for x in args] # type: ignore[name-defined]
        return cls._build_decorator(keys)

    @classmethod
    def references(cls, *args, **kwargs) -> Decorator:
        """ mark keys to use as to_coderef imports """
        keys = [DKey[CodeReference](x, implicit=True, **kwargs) for x in args] # type: ignore[name-defined]
        return cls._build_decorator(keys)

    @classmethod
    def postbox(cls, *args, **kwargs) -> Decorator:
        keys = [DKey[DKey.Marks.POSTBOX](x, implicit=True, **kwargs) for x in args] # type: ignore[name-defined]
        return cls._build_decorator(keys)



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports


# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from ._interface import Logger
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class LogContext:
    """
      a really simple wrapper to set a logger's level, then roll it back

    use as:
    with LogContext(logger, level=logmod.INFO) as ctx:
    ctx.log("blah")
    # or
    logger.info("blah")
    """

    def __init__(self, logger, level=None):
        self._logger          = logger
        self._original_level  = self._logger.level
        self._level_stack     = [self._original_level]
        self._temp_level      = level or self._original_level

    def __call__(self, level) -> Self:
        self._temp_level = level
        return self

    def __enter__(self) -> Self:
        match self._temp_level:
            case int() | str():
                self._level_stack.append(self._logger.level)
                self._logger.setLevel(self._temp_level)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        if bool(self._level_stack):
            self._logger.setLevel(self._level_stack.pop())
        else:
            self._logger.setLevel(self._original_level)

    def __getattr__(self, key):
        return getattr(self._logger, key)

    def log(self, msg, *args, **kwargs):
        self._logger.log(self._temp_level, msg, *args, **kwargs)

class TempLogger:
    """ For using a specific type of logger in a context, or getting
    a custom logger class without changing it globally

    use as:
    with TempLogger(MyLoggerClass) as ctx:
    # Either:
    ctx['name'].info(...)
    # or:
    logmod.getLogger('name').info(...)
    """

    def __init__(self, logger:type[Logger]):
        self._target_cls                        = logger
        self._original : Maybe[logmod.Logger]   = None

    def __enter__(self) -> Self:
        self._original = logmod.getLoggerClass()
        logmod.setLoggerClass(self._target_cls)
        return self

    def __exit(self, *exc):
        self.setLoggerClass(self._original)
        return False



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

# Vars:
DEL_LOG_K         : Final[str] = "_log_del_active"
DEFAULT_LOG_LEVEL : Final[int] = 40
DEFAULT_REPORT    : Final[str] = "traceback"
CHANGE_PREFIX     : Final[str] = "+-:"
DEFAULT_PREFIX    : Final[str] = "--"
# Body:



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, MaybeT
    type InstanceData      = dict
    type PostInstanceData  = dict

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

type HookOverride         = bool
type PreProcessResult[T]  = tuple[str, InstanceData, PostInstanceData, Maybe[type[T]]]
# Body:

class PreProcessor_p[T](Protocol):
    """ Protocol for things like Strang,
    whose metaclass preprocess the initialisation data before even __new__ is called.

    Is used in a metatype.__call__ as::

        cls._pre_process(...)
        obj = cls.__new__(...)
        obj.__init__(...)
        obj._process()
        obj._post_process()
        return obj

    """

    def pre_process(self, cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> PreProcessResult[T]: ...  # noqa: A002, ANN401

    def process(self, obj:T, *, data:PostInstanceData) -> Maybe[T]: ...

    def post_process(self, obj:T, *, data:PostInstanceData) -> Maybe[T]: ...

class ProcessorHooks[T](Protocol):
    """ Hooks a PreProcessor_p recognizes.

    returning True as the first result value means the processor's logic is *not* to be used.
    otherwise it *is* used, with the results from the hook
    """

    @classmethod
    def _pre_process_h(cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> Maybe[tuple[bool, *PreProcessResult[T]]]: ...  # noqa: A002, ANN401

    def _process_h(self, *, data:PostInstanceData) -> Maybe[tuple[HookOverride, Maybe[T]]]: ...

    def _post_process_h(self, *, data:PostInstanceData) -> Maybe[tuple[HookOverride, Maybe[T]]]: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

from dataclasses import dataclass, field, InitVar
from typing import Any

if TYPE_CHECKING:
    import types
    from jgdv import Maybe, Rx
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
DEFAULT_PREFIX  : Final[str]  = "-"
END_SEP         : Final[str]  = "--"
FULLNAME_RE     : Final[Rx]   = re.compile(r"(?:<(?P<pos>\d*)>|(?P<prefix>\W+))?(?P<name>.+?)(?P<assign>=)?$")
DEFAULT_DOC     : Final[str]  = "A Base Parameter"
""" The Regexp for parsing string descriptions of parameters """

##--|
EMPTY_CMD           : Final[str]  = "_cmd_"
EXTRA_KEY           : Final[str]  = "_extra_"
NON_DEFAULT_KEY     : Final[str]  = "_non_default_"
DEFAULT_COUNT       : Final[int]  = 1
UNRESTRICTED_COUNT  : Final[int]  = -1
##--|
TYPE_CONV_MAPPING: Final[dict[str|type|types.GenericAlias, type|Callable]] = {
    "int"               : int,
    "float"             : float,
    "bool"              : bool,
    "str"               : str,
    "list"              : list,
}

# Body:

class ParseResult_d:
    __slots__ = ("args", "name", "non_default")
    name        : str
    args        : dict
    non_default : set[str]

    def __init__(self, name:str, args:Maybe[dict]=None, non_default:Maybe[set]=None) -> None:
        self.name         = name
        self.args         = args or {}
        self.non_default  = non_default or set()

    def __repr__(self) -> str:
        return f"<ParseResult: {self.name}, args:{self.args}>"

    def to_dict(self) -> dict:
        return {"name":self.name, "args":self.args, NON_DEFAULT_KEY:self.non_default}
##--| Params

@runtime_checkable
class ParamStruct_p(Protocol):
    """ Base class for CLI param specs, for type matching
    when 'consume' is given a list of strs,
    it can match on the args,
    and return an updated diction and a list of values it didn't consume

    """

    @classmethod
    def key_func(cls, x:ParamStruct_i) -> tuple: ...

    def consume(self, args:list[str], *, offset:int=0) -> Maybe[tuple[dict, int]]:
        pass

    ##--| properties

    @property
    def short(self) -> str: ...

    @property
    def inverse(self) -> str: ...

    @property
    def repeatable(self) -> bool: ...

    @property
    def key_str(self) -> str: ...

    @property
    def short_key_str(self) -> Maybe[str]: ...

    @property
    def key_strs(self) -> list[str]: ...

    @property
    def default_value(self) -> Any: ...  # noqa: ANN401

    @property
    def default_tuple(self) -> tuple[str, Any]: ...

class ParamStruct_i(ParamStruct_p, Protocol):
    _processor : ClassVar

    name       : str
    type_      : Maybe[type]
    insist     : bool
    default    : Any|Callable
    desc       : str
    count      : int
    prefix     : int|str
    separator  : str|Literal[False]
    implicit   : bool

##--| Param Subtypes

@runtime_checkable
class PositionalParam_p(Protocol):

    def _positional(self) -> Literal[True]: ...

@runtime_checkable
class AssignmentParam_p(Protocol):

    def _assignment(self) -> Literal[True]: ...

@runtime_checkable
class KeyParam_p(Protocol):

    def _keyval(self) -> Literal[True]: ...

@runtime_checkable
class ToggleParam_p(Protocol):

    def _toggle(self) -> Literal[True]: ...
##--| Parsing

@runtime_checkable
class ArgParser_p(Protocol):
    """
    A Single standard process point for turning the list of passed in args,
    into a dict, into a chainguard,
    along the way it determines the cmds and tasks that have been chosne
    """

    def _parse_fail_cond(self) -> bool:
        raise NotImplementedError()

    def _has_no_more_args_cond(self) -> bool:
        raise NotImplementedError()

@runtime_checkable
class ParamSource_p(Protocol):
    """ Param Sources are anything that can provide a name and a set of parameters """

    @property
    def name(self) -> str:
        raise NotImplementedError()

    def param_specs(self) -> list[ParamStruct_i]:
        raise NotImplementedError()

@runtime_checkable
class CLIParamProvider_p(Protocol):
    """
      Things that can provide parameter specs for CLI parsing
    """

    @classmethod
    def param_specs(cls) -> list[ParamStruct_i]:
        """  make class parameter specs  """
        pass



#!/usr/bin/env python3
"""
TEST File updated

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

##--|
from .. import _interface as API # noqa: N812

##--|

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestSec_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match API.Sec_d("test", ".", "::", str|int, API.DefaultBodyMarks_e, True, idx=0):
            case API.Sec_d():
                assert(True)
            case x:
                assert(False), x

    def test_contains_mark(self):
        obj = API.Sec_d("test", ".", "::", str|int, API.DefaultBodyMarks_e, True, idx=0)
        assert(API.DefaultBodyMarks_e.head in obj)

    def test_contains_mark_fail(self):
        obj = API.Sec_d("test", ".", "::", str|int, API.DefaultBodyMarks_e, True, idx=0)
        assert(API.DefaultHeadMarks_e.basic not in obj)

    def test_contains_type(self):
        obj = API.Sec_d("test", ".", "::", str|int|list, API.DefaultBodyMarks_e, True, idx=0)
        assert(str in obj)

    def test_contains_type_fail(self):
        obj = API.Sec_d("test", ".", "::", str|int|list, API.DefaultBodyMarks_e, True, idx=0)
        assert(float not in obj)

    def test_contains_type_union(self):
        obj = API.Sec_d("test", ".", "::", str|int|list, API.DefaultBodyMarks_e, True, idx=0)
        assert(str|int in obj)

    def test_contains_type_union_fail(self):
        obj = API.Sec_d("test", ".", "::", str|int|list, API.DefaultBodyMarks_e, True, idx=0)
        assert(str|int|float not in obj)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from . import _interface as API # noqa: N812

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from types import GenericAlias
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
type AliasAnnotation = type|str|enum.Enum|Literal[False]|Literal[True]|tuple[AliasAnnotation, ...]

# Body:

class SubAlias_m:
    """ A Mixin to manage generics that resolve to specific registered subclasses.

    On class declaration, recognizes kwargs:
    - fresh_registry:bool                         : use a separate registry for this class and subclasses
    - accumulate:bool                             : annotations accumulate from their parent class
    - strict:bool                                 : error if a subclass tries to overwrite a registration
    - default:bool                                : set this subclass as the default if no marks are specified when creating an instance
    - annotation:Maybe[str|type|enum|tuple[...]]  : the key to use for this subclass, if class_getitem hasn't been used
    - no_register:bool                            : create the class, but don't register it

    cls[val] -> GenericAlias(cls, val)

    then:

    class RealSub(cls[val]) ...
    after which:
    cls[val] is RealSub

    Annotation Keys are stored in cls.__annotation__,
    under the cls._annotate_to key name.

    """
    __slots__                                              = ()
    _annotate_to  : ClassVar[str]                          = API.AnnotationTarget
    _default_k    : ClassVar[str]                          = API.Default_K
    # TODO make this a weakdict?
    _registry     : ClassVar[dict[AliasAnnotation, type]]  = {}
    _strict       : ClassVar[bool]                         = False
    _accumulator  : ClassVar[bool]                         = False

    def __init_subclass__(cls:type[Self], *args:Any, annotation:Maybe[AliasAnnotation]=None, fresh_registry:bool=False, **kwargs:Any) -> None:  # noqa: ANN401
        x  : Any
        overwrite : bool
        # ensure a new annotations dict
        # (if a subclass doesn't add *new* annotations, the super's is used. so if we modify it here,
        # the subclass is effected)
        cls.__annotations__  = cls.__annotations__.copy()
        overwrite            = kwargs.pop("overwrite", False)
        if (strict:=kwargs.pop("strict", None)):
            cls._strict = strict or cls._strict
        if (accumulate:=kwargs.pop("accumulate", None)):
            cls._accumulator = accumulate or cls._accumulator

        if fresh_registry:
            cls._registry    = {}

        if kwargs.pop("default", None) is True:
            cls._registry[cls._default_k] = cls

        # set the annotation target
        match kwargs.pop(API.AnnotateKWD, None):
            case str() as target:
                logging.debug("Annotate Subclassing: %s : %s", cls, kwargs)
                cls._annotate_to               = target
                setattr(cls, cls._annotate_to, None)
            case None if not hasattr(cls, cls._annotate_to):
                setattr(cls, cls._annotate_to, None)
            case _:
                pass

        annotation                             = cls._build_annotation(annotation)
        cls.__annotations__[cls._annotate_to]  = annotation
        if kwargs.pop("no_register", False):
            return

        match annotation, cls._registry.get(annotation, None):
            case (), _: # No annotation
                pass
            case _, None: # No registered cls
                cls._registry[annotation] = cls
            case _, x if overwrite:
                cls._registry[annotation] = cls
            case _, x if cls._strict: # complain there s a cls
                msg = "already has a registration"
                raise TypeError(msg, x, cls, annotation, args, kwargs)

    def __class_getitem__[K:AliasAnnotation](cls:type[Self], *key:K) -> type|GenericAlias:
        return cls._retrieve_subtype(*key)

    @classmethod
    def _retrieve_subtype[K:AliasAnnotation](cls:type[Self], key:K) -> type|GenericAlias:
        use_key : AliasAnnotation = cls._build_annotation(key)
        match cls._registry.get(use_key, None):
            case type() as result:
                return result
            case None if use_key == () and cls._default_k in cls._registry:
                return cls._registry[cls._default_k]
            case _:
                return GenericAlias(cls, use_key)

    @classmethod
    def _clear_registry(cls) -> None:
        cls._registry.clear()

    @classmethod
    def cls_annotation(cls) -> tuple:
        return cls.__annotations__.get(cls._annotate_to, ())

    @classmethod
    def _build_annotation(cls, annotate:Maybe[AliasAnnotation]) -> AliasAnnotation:  # noqa: PLR0912
        """ single point of truth for determining annotations from a cls and provided annotation """
        x : Any
        result : list[AliasAnnotation] = []
        assert(hasattr(cls, "cls_annotation"))

        match cls.cls_annotation():
            case []:
                pass
            case tuple() as x:
                result += x

        match cls.__dict__.get(API.ORIG_BASES_K, []):
            case _ if bool(result):
                pass
            case [GenericAlias() as x]:
                result += x.__args__
            case _ if (annotated:=[x for x in cls.mro() if x is not cls and hasattr(x, "cls_annotation")]):
                result += annotated[0].cls_annotation()
            case _:
                pass

        match annotate:
            case None:
                pass
            case [[*xs]] if cls._accumulator:
                result += xs
            case tuple() as x if cls._accumulator:
                result += x
            case x if cls._accumulator:
                result.append(x)
            case [[*xs]]:
                result = [*xs]
            case [*xs]:
                result = [*xs]
            case x:
                result = [x]

        ##--|
        match result:
            case []:
                return ()
            case [x]:
                return tuple([x])
            case [*xs]:
                return tuple(xs)
            case _:
                return ()



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

logging = logmod.root

from jgdv.files.tags.tag_file import TagFile
from jgdv.files.tags.sub_file import SubstitutionFile

class TestSubFile:

    def test_initial(self):
        obj = SubstitutionFile()
        assert(obj is not None)

    def test_len(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(len(obj) == 3)

    def test_sub_default(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(obj.sub("a") == {"a"})

    def test_sub_norms(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(obj.sub("a tag") == {"a_tag"})
        assert(obj.sub("a_tag") == {"a_tag"})

    def test_update_just_count(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(obj.get_count("b") == 5)
        obj.update("b")
        assert(obj.get_count("b") == 6)

    def test_update_subs(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(obj.sub("a_tag") == {"a_tag"})
        obj.update(("a tag", 1, "blah"))
        assert(obj.sub("a_tag") == {"blah"})


    def test_contains(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert("a tag" in obj)
        assert("a_tag" in obj)


    def test_contains_subs(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert("a tag" in obj)
        assert("a_tag" in obj)
        obj.update(("a tag", 1, "blah"))
        assert("blah" in obj)


    def test_subs_dont_have_subs(self):
        obj = SubstitutionFile(count={"a": 2, "b": 5, "a tag": 19})
        obj.update(("a tag", 1, "blah"))
        assert("blah" in obj)
        assert(not obj.has_sub("blah"))


    def test_has_sub_false(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(not obj.has_sub("a"))
        assert(not obj.has_sub("a_tag"))

    def test_has_sub_on_norm(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(obj.has_sub("a tag"))
        assert(not obj.has_sub("a_tag"))

    def test_has_sub_true(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(not obj.has_sub("a_tag"))
        obj.update(("a_tag", 1, "blah"))
        assert(obj.has_sub("a tag"))


    def test_update_multi_sub(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        assert(not obj.has_sub("a_tag"))
        obj.update(("a_tag", 1, "blah", "bloo"))
        assert(obj.sub("a_tag") == {"blah", "bloo"})


    def test_canonical(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        obj.update(("a_tag", 1, "blah", "bloo"))
        canon = obj.canonical()
        assert(isinstance(canon, TagFile))
        assert("a" in canon)
        assert("b" in canon)
        assert("blah" in canon)
        assert("bloo" in canon)


    def test_canonical_filters_presubs(self):
        obj = SubstitutionFile(counts={"a": 2, "b": 5, "a tag": 19})
        obj.update(("a_tag", 1, "blah", "bloo"), ("bloo", 1, "aweg"))
        canon = obj.canonical()
        assert(isinstance(canon, TagFile))
        assert("a_tag" not in canon)
        assert("bloo" not in canon)
        assert("aweg" in canon)




#!/usr/bin/env python3
"""

"""
# ruff: noqa: N801
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    import pathlib as pl
    import enum
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from .types import Maybe, Ctor
    type ChainGuard = Any
    type Logger     = logmod.Logger

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@runtime_checkable
class ArtifactStruct_p(Protocol):
    """ Base class for artifacts, for type matching """

    def exists(self, *, data=None) -> bool: ...  # noqa: ANN001

@runtime_checkable
class UpToDate_p(Protocol):
    """ For things (often artifacts) which might need to have actions done if they were created too long ago """

    def is_stale(self, *, other:Any=None) -> bool: ...  # noqa: ANN401
    """ Query whether the task's artifacts have become stale and need to be rebuilt"""

@runtime_checkable
class StubStruct_p(Protocol):
    """ Base class for stubs, for type matching """

    def to_toml(self) -> str: ...

@runtime_checkable
class SpecStruct_p(Protocol):
    """ Base class for specs, for type matching """

    @property
    def params(self) -> dict|ChainGuard: ...

    @property
    def args(self) -> list: ...

    @property
    def kwargs(self) -> dict: ...

@runtime_checkable
class TomlStubber_p(Protocol):
    """
      Something that can be turned into toml
    """

    @classmethod
    def class_help(cls) -> str: ...

    @classmethod
    def stub_class(cls, stub:StubStruct_p) -> None: ...
    """
        Specialize a StubStruct_p to describe this class
    """

    def stub_instance(self, stub:StubStruct_p) -> None: ...
    """
        Specialize a StubStruct_p with the settings of this specific instance
    """

    @property
    def short_doc(self) -> str: ...
    """ Generate Job Class 1 line help string """

    @property
    def doc(self) -> list[str]: ...

@runtime_checkable
class ActionGrouper_p(Protocol):
    """ For things have multiple named groups of actions """

    def get_group(self, name:str) -> Maybe[list]: ...

@runtime_checkable
class Loader_p(Protocol):
    """ The protocol for something that will load something from the system, a file, etc
    TODO add a type parameter
    """

    def setup(self, extra_config:ChainGuard) -> Self: ...

    def load(self) -> ChainGuard: ...

@runtime_checkable
class Buildable_p(Protocol):
    """ For things that need building, but don't have a separate factory
    TODO add type parameter
    """

    @classmethod
    def build(cls, *args:Any) -> Self: ...  # noqa: ANN401

@runtime_checkable
class Factory_p[T](Protocol):
    """
      Factory protocol: {type}.build
    """

    @classmethod
    def build(cls:Ctor[T], *args:Any, **kwargs:Any) -> T: ...  # noqa: ANN401

@runtime_checkable
class Nameable_p(Protocol):
    """ The core protocol of something use as a name """

    def __hash__(self) -> int: ...

    def __eq__(self, other:object) -> bool: ...

    def __lt__(self, other:Nameable_p) -> bool: ...

    def __contains__(self, other:Nameable_p) -> bool: ...

@runtime_checkable
class InstantiableSpecification_p(Protocol):
    """ A Specification that can be instantiated further """

    def instantiate_onto(self, data:Maybe[Self]) -> Self: ...

    def make(self) -> Self: ...

@runtime_checkable
class ExecutableTask(Protocol):
    """ Runners pass off to Tasks/Jobs implementing this protocol
      instead of using their default logic
    """

    def setup(self) -> None: ...
    """ """

    def expand(self) -> list: ...
    """ For expanding a job into tasks """

    def execute(self) -> None: ...
    """ For executing a task """

    def teardown(self) -> None: ...
    """ For Cleaning up the task """

    def check_entry(self) -> bool: ...
    """ For signifiying whether to expand/execute this object """

    def execute_action_group(self, group_name:str) -> enum.Enum|list: ...
    """ Optional but recommended """

    def execute_action(self) -> None: ...
    """ For executing a single action """

    def current_status(self) -> enum.Enum: ...

    def force_status(self, status:enum.Enum) -> None: ...

    def current_priority(self) -> int: ...

    def decrement_priority(self) -> None: ...

@runtime_checkable
class Persistent_p(Protocol):
    """ A Protocol for persisting data """

    def write(self, target:pl.Path) -> None: ...
    """ Write this object to the target path """

    def read(self, target:pl.Path) -> None: ...
    """ Read the target file, creating a new object """

@runtime_checkable
class FailHandler_p(Protocol):

    def handle_failure(self, err:Exception, *args:Any, **kwargs:Any) -> Maybe[Any]: ...  # noqa: ANN401

@runtime_checkable
class Visitor_p(Protocol):

    def visit(self, **kwargs:Any) -> Any: ...  # noqa: ANN401

@runtime_checkable
class DILogger_p(Protocol):
    """ Protocol for classes with a dependency injectable logger """

    def logger(self) -> Logger: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from . import _interface as API # noqa: N812

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from types import GenericAlias, resolve_bases
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never, TypeAliasType
from typing import no_type_check, final, override, overload, _caller # type: ignore[attr-defined]
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable

from pydantic import BaseModel, create_model

if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class Subclasser:
    """ A Util class for building subclasses programmatically

    Subclasses can have modified mro's,
    Also extended namespaces,
    And preserve the base class' __slots__/__dict__ state

    """

    @staticmethod
    def decorate_name(cls:str|type, *vals:str, params:Maybe[str]=None) -> Maybe[str]:  # noqa: PLW0211
        """ Create a new name for an annotated subclass

        decorate(cls, a,b,c) -> cls<+a+b+c>
        decorate(cls, params='blah') -> cls[blah]
        """
        name         : str
        annotations  : Maybe[str]  = None
        set_extras   : set[str]    = set(vals)

        match cls:
            case x if not (bool(params) or bool(vals)):
                return None
            case type() as x:
                name = x.__name__
            case str() as x:
                name = x
            case x:
                raise TypeError(API.BadDecorationNameTarget, x)

        match API.AnnotateRx.match(name):
            case re.Match() as mtch:
                set_extras.update({y for y in (mtch['extras'] or "").split("+") if bool(y)})
                params  = params or mtch['params'] or None
                name    = mtch['name'] or name
            case _:
                raise ValueError(API.NoNameMatch, cls)

        if bool(set_extras):
            annotations = "+".join(x for x in sorted(set_extras) if bool(x))

        match annotations, params:
            case str() as x, None:
                return f"{name}<+{x}>"
            case None, str() as x:
                return f"{name}[{params}]"
            case str() as x, str() as y:
                return f"{name}<+{x}>[{y}]"
            case _:
                return None

    def annotate[T](self, cls:type[T], *params:Any) -> type[T]:  # noqa: ANN401
        """ Make a subclass of cls,

        annotated to have params in getattr(cls, '_annotate_to', '_typevar')
        """
        p_str        : str
        def_mod      : str
        subname      : str
        namespace    : dict
        anno_target  : str  = getattr(cls, API.AnnotateKWD, API.AnnotationTarget)
        anno_type    : str  = "ClassVar[str]"
        match params:
            case [NewType() as param]:
                p_str = param.__name__  # type: ignore[attr-defined]
            case [TypeAliasType() as param]:
                p_str = param.__value__.__name__
            case [type() as param]:
                p_str = param.__name__
            case [str() as param]:
                p_str = param
            case [param]:
                p_str = str(param)
            case [_, *_]:  # type: ignore[misc]
                raise NotImplementedError(API.MultiParamFail, params)
            case _:
                raise ValueError(API.BadParamFail, params)

        # Get the module definer 3 frames up.
        # So not annotate, or __class_getitem__, but where the subclass is created
        def_mod = _caller(3)
        match self.decorate_name(cls, params=p_str):
            case str() as x:
                subname   = x
                namespace = {
                    anno_target : param,
                    API.MODULE_NAME : def_mod,
                    API.ANNOTS_NAME : {anno_target : anno_type},
                }
                sub = self.make_subclass(subname, cls, namespace=namespace)
                setattr(sub, anno_target , param)  # type: ignore[attr-defined]
                return sub
            case _:
                raise ValueError(API.NoSubName)

    def make_generic[T](self, cls:type[T], *params:Any) -> GenericAlias:  # noqa: ANN401
        return GenericAlias(cls, *params)

    def make_subclass[T](self, name:str, cls:type[T], *, namespace:Maybe[dict]=None, mro:Maybe[Iterable]=None) -> type[T]:
        """
        Build a dynamic subclass of cls, with name,
        possibly with a maniplated mro and internal namespace
        """
        if (ispydantic:=issubclass(cls, BaseModel)) and mro is not None:
            raise NotImplementedError(API.NoPydanticFail)
        elif ispydantic:
            sub = self._new_pydantic_class(name, cls, namespace=namespace)
            return sub
        else:
            sub = self._new_std_class(name, cls, namespace=namespace, mro=mro)
            return sub

    def _new_std_class[T](self, name:str, cls:type[T], *, namespace:Maybe[dict]=None, mro:Maybe[Iterable]=None) -> type[T]:
        """
        Dynamically creates a new class
        """
        from types import new_class
        assert(not issubclass(cls, BaseModel)), cls
        mod_name  : str
        mcls      : type[type]  = type(cls)
        # if name == cls.__name__:
        #     name = f"{name}<+>"

        match namespace:
            case dict():
                pass
            case _:
                namespace = {}
        match mro:
            case None:
                mro = cls.mro()
            case tuple() | list():
                pass
            case x:
                raise TypeError(API.UnexpectedMRO, x)
        ##--|
        assert(namespace is not None)
        # Expand out generics by calling __mro_entries__
        match (mro:=tuple(resolve_bases(mro))):
            case [x, *_]: # Use the base class module name
                mod_name = x.__dict__[API.MODULE_NAME]
                namespace.setdefault(API.MODULE_NAME, mod_name)
            case _:
                raise ValueError()

        # mro_dict = ChainMap(*(getattr(c, '__dict__', {}) for c in cls.mro()))
        namespace.setdefault(API.SLOTS_NAME, ())
        try:
            return mcls(name, mro, namespace)
        except TypeError as err:
            err.add_note(str(mro))
            raise

    def _new_pydantic_class(self, name:str, cls:type, *, namespace:Maybe[dict]=None) -> type:
        assert(issubclass(cls, BaseModel)), cls
        sub = create_model(name, __base__=cls)
        for x,y in (namespace or {}).items():
            setattr(sub, x, y)
        return sub



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import builtins
import datetime
import enum
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv.mixins.annotate.annotate import SubAnnotate_m
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from jgdv.cli.errors import ArgParseError
from .param_spec import ParamSpec
from .core import ToggleParam
from .extra import RepeatToggleParam, RepeatToggleParam, LiteralParam

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
from .._interface import ParamStruct_p
# isort: on
# ##-- end types
##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


class HelpParam(ToggleParam): #[bool]):
    """ The --help flag that is always available """

    desc : str = "The Default Help Param"

    def __init__(self, **kwargs):
        kwargs.update({"name":"--help", "default":False, "implicit":True})
        super().__init__(**kwargs)

class VerboseParam(RepeatToggleParam): #[int]):
    """ The implicit -verbose flag """

    desc : str = "The Default Verbosity Param"

    def __init__(self, **kwargs):
        kwargs.update({"name":"--verbose", "default":0, "implicit":True})
        super().__init__(**kwargs)

class SeparatorParam(LiteralParam):
    """ A Parameter to separate subcmds """

    desc : str = "The Default Separator Param"

    def __init__(self, **kwargs):
        kwargs.update({"name":"--", "implicit":True})
        super().__init__(**kwargs)



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

from . import _interface as API # noqa: N812
from ._interface import INDIRECT_SUFFIX, RAWKEY_ID, DKeyMark_e
from ._util._interface import ExpInst_d, SourceChain_d
from .dkey import DKey

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, Self
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from typing import Never
from typing import Any
from collections.abc import Mapping

if TYPE_CHECKING:
   import pathlib as pl
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, LiteralString
   from typing import Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, MutableMapping, Hashable
   from jgdv._abstract.protocols import SpecStruct_p

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging
class SingleDKey(DKey, mark=Any, default=True):
    """
      A Single key with no extras.
      ie: {x}. not {x}{y}, or {x}.blah.
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        super().__init__(*args, **kwargs)
        match self.data.raw:
            case [x] if self.data.convert is None:
                self.data.convert  = x.convert
                self.data.format   = x.format
            case [_]:
                pass
            case None | []:
                msg = "A Single Key has no raw key data"
                raise ValueError(msg)
            case [*xs]:
                msg = "A Single Key got multiple raw key data"
                raise ValueError(msg, xs)

    def _post_process_h(self, data:Maybe[dict]=None) -> None:  # noqa: ARG002
        match self.data.raw[0].format:
            case None:
                return
            case str() as format:
                pass

        match API.EXPANSION_LIMIT_PATTERN.search(format):
            case re.Match() as m:
                self.data.max_expansions = int(m[1])
            case None:
                pass

class MultiDKey(DKey, mark=list):
    """ Multi keys allow 1+ explicit subkeys.

    They have additional fields:

    _subkeys  : parsed information about explicit subkeys

    """
    __slots__ = ("anon",)
    anon  : str

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        super().__init__(*args, **kwargs)
        match self.data.raw:
            case [] | None:
                msg = "Tried to build a multi key with no subkeys"
                raise ValueError(msg, self.data.raw, kwargs)
            case [*xs]:
                self.anon  = "".join(x.anon() for x in xs)

    def __str__(self) -> str:
        return self[:]

    def __contains__(self, other:object) -> bool:
         return other in self.keys()

    def __format__(self, spec:str, **kwargs:Any) -> str:  # noqa: ANN401
        """ Just does normal str formatting """
        rem, _, _= self._processor.consume_format_params(spec) # type: ignore
        return super().__format__(rem, **kwargs)

    def _multi(self) -> Literal[True]:
        return True

    def keys(self) -> list[DKey]: # type: ignore[override]
        return [DKey(x, force=DKey[Any], implicit=True) for x in self.data.meta if bool(x)]

    def exp_generate_alternatives_h(self, sources:SourceChain_d, opts:dict) -> list[list[ExpInst_d]]:  # noqa: ARG002
        """ Lift subkeys to expansion instructions """
        targets = []
        for key in self.keys():
            targets.append([ExpInst_d(value=key,
                                      convert=key.data.convert,
                                      rec=key.data.max_expansions,
                                      fallback=None),
                            ])
        else:
            return targets

    def exp_flatten_h(self, vals:list[ExpInst_d], opts:dict) -> Maybe[ExpInst_d]:  # noqa: ARG002
        """ Flatten the multi-key expansion into a single string,
        by using the anon-format str
        """
        flat : list[str]  = []
        key_meta          = [x for x in self.data.meta if bool(x)]
        if not bool(key_meta):
            return vals[0]
        if len(vals) != len([x for x in self.data.meta if bool(x)]):
            return None

        for x in vals:
            match x:
                case ExpInst_d(value=IndirectDKey() as k):
                    flat.append(f"{k:wi}")
                case ExpInst_d(value=API.Key_p() as k):
                    flat.append(f"{k:w}")
                case ExpInst_d(value=x):
                    flat.append(str(x))
        else:
            return ExpInst_d(value=self.anon.format(*flat), literal=True)

class NonDKey(DKey, mark=False):
    """ Just a string, not a key.

    ::

        But this lets you call no-ops for key specific methods.
        It can coerce itself though
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        super().__init__(*args, **kwargs)
        if (fb:=kwargs.get('fallback', None)) is not None and fb != self:
            msg = "NonKeys can't have a fallback, did you mean to use an explicit key?"
            raise ValueError(msg, self)

    def __format__(self, spec:str, **kwargs:Any) -> str:  # noqa: ANN401
        """ Just does normal str formatting """
        rem, _, _= self._processor.consume_format_params(spec) # type: ignore
        return super().__format__(rem, **kwargs)

    def _nonkey(self) -> Literal[True]:
        return True

    def expand(self, *args, **kwargs) -> Maybe:  # noqa: ANN002, ANN003, ARG002
        """ A Non-key just needs to be coerced into the correct str format """
        assert(isinstance(self, API.Key_p))
        val = ExpInst_d(value=self[:])
        match self._expander.coerce_result(val, kwargs, source=self):
            case None if (fallback:=kwargs.get("fallback")) is not None:
                return ExpInst_d(value=fallback, literal=True)
            case None:
                return self.data.fallback
            case ExpInst_d() as x:
                return x.value
            case x:
                msg = "Nonkey coercion didn't return an ExpInst_d"
                raise TypeError(msg, x)

class IndirectDKey(DKey, mark=Mapping, convert="I"):
    """
      A Key for getting a redirected key.
      eg: RedirectionDKey(key) -> SingleDKey(value)

      re_mark :
    """
    __slots__  = ("multi_redir", "re_mark")
    __hash__ = SingleDKey.__hash__

    def __init__(self, *, multi:bool=False, re_mark:Maybe[API.KeyMark]=None, **kwargs) -> None:  # noqa: ANN003
        assert(not self.endswith(INDIRECT_SUFFIX)), self[:]
        kwargs.setdefault("fallback", Self)
        super().__init__(**kwargs)
        self.multi_redir      = multi
        self.re_mark          = re_mark

    def __eq__(self, other:object) -> bool:
        match other:
            case str() if other.endswith(INDIRECT_SUFFIX):
                return f"{self:i}" == other
            case _:
                return super().__eq__(other)

    def _indirect(self) -> Literal[True]:
        return True

    def exp_generate_alternatives_h(self, sources:list[dict], opts:dict) -> list[list[ExpInst_d]]:  # noqa: ARG002
        """ Lookup the indirect version, the direct version, then use the fallback """
        match opts.get("fallback", self.data.fallback):
            case x if x is Self:
                fallback = self
            case None:
                fallback = self
            case x:
                fallback = x
        return [[
            ExpInst_d(value=f"{self:i}", lift=True),
            ExpInst_d(value=f"{self:d}", convert=False),
            ExpInst_d(value=fallback,    literal=True, convert=False),
        ]]



#!/usr/bin/env python3
"""
see `Alexandra Zaharia <https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging/>`_

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging
import logging as logmod
import pathlib as pl
import re
import warnings
from collections import defaultdict
from string import Formatter
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import _string
import sty
from sty import bg, ef, fg, rs

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Mixin, Proto

# ##-- end 1st party imports

from . import _interface as API # noqa: N812
from .stack_m import StackFormatter_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Rx
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from logging import LogRecord

    type StyleChar = Literal["%", "{", "$"]
##--|

# isort: on
# ##-- end types

COLOUR_RESET       : str    = rs.all
##--|


@Mixin(StackFormatter_m)
class ColourFormatter(logging.Formatter):
    """
    Stream Formatter for logging, enables use of colour sent to console

    Guarded Formatter for adding colour.
    Uses the sty module.
    If sty is missing, behaves as the default formatter class

    # Do *not* use for on filehandler
    Usage reminder:
    # Create stdout handler for logging to the console (logs all five levels)
    stdout_handler = logging.StreamHandler()
    stdout_handler.setFormatter(ColourFormatter(fmt))
    logger.addHandler(stdout_handler)
    """

    _default_fmt      : str        = '{asctime} | {levelname:9} | {message}'
    _default_date_fmt : str        =  "%H:%M:%S"
    _default_style    : StyleChar  = '{'
    colours           : dict[int|str, str]

    def __init__(self, *, fmt:Maybe[str]=None, style:Maybe[StyleChar]=None) -> None:
        """
        Create the ColourFormatter with a given *Brace* style log format
        """
        super().__init__(fmt or self._default_fmt,
                         datefmt=self._default_date_fmt,
                         style=style or self._default_style)
        self.colours = defaultdict(lambda: rs.all)
        self.apply_colour_mapping(API.default_colour_mapping)
        self.apply_colour_mapping(API.default_log_colours)

    def format(self, record:LogRecord) -> str:
        if hasattr(record, "colour"):
            log_colour = self.colours[record.colour]
        else:
            log_colour = self.colours[record.levelno]

        return log_colour + super().format(record) + COLOUR_RESET

    def apply_colour_mapping(self, mapping:dict) -> None:
        """ applies a mapping of colours by treating each value as a pair of attrs of sty

        eg: {logging.DEBUG: ("fg", "blue"), logging.INFO: ("bg", "red")}
        """
        for x,(a,b) in mapping.items():
            accessor = getattr(sty, a)
            val      = getattr(accessor, b)
            self.colours[x] = val


@Mixin(StackFormatter_m)
class StripColourFormatter(logging.Formatter):
    """
    Force Colour Command codes to be stripped out of a string.
    Useful for when you redirect printed strings with colour
    to a file
    """

    _default_fmt      : str           = "{asctime} | {levelname:9} | {shortname:25} | {message}"
    _default_date_fmt : str           = "%Y-%m-%d %H:%M:%S"
    _default_style    : StyleChar     = '{'
    _colour_strip_re  : Rx            = re.compile(r'\x1b\[([\d;]+)m?')

    def __init__(self, *, fmt:Maybe[str]=None, style:Maybe[StyleChar]=None) -> None:
        """
        Create the StripColourFormatter with a given *Brace* style log format
        """
        super().__init__(fmt or self._default_fmt,
                         datefmt=style or self._default_date_fmt,
                         style=self._default_style)

    def format(self, record:LogRecord) -> str:
        result    = super().format(record)
        no_colour = self._colour_strip_re.sub("", result)
        return no_colour



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
import _string
# ##-- end stdlib imports

from .. import _interface as API # noqa: N812

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Ident
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
OBRACE : Final[str] = "{"
CBRACE : Final[str] = "}"
# Body:

class DKeyParser:
    """ Parser for extracting {}-format params from strings.

    ::

        see: https://peps.python.org/pep-3101/
        and: https://docs.python.org/3/library/string.html#format-string-syntax
    """

    def parse(self, format_string:str, *, implicit:bool=False) -> Generator[API.RawKey_d]:
        if implicit and OBRACE in format_string:
            msg = "Implicit key already has braces"
            raise ValueError(msg, format_string)

        if implicit: # Wrap implicit keys with braces, to extract format and conv parameters
            format_string = "".join([OBRACE, format_string, CBRACE])  # noqa: FLY002

        try:
            for x in _string.formatter_parser(format_string):
                yield self.make_param(*x)
        except ValueError:
            yield self.make_param(format_string)

    def make_param(self, prefix:str, key:Maybe[str]=None, format:Maybe[str]=None, convert:Maybe[str]=None) -> API.RawKey_d:  # noqa: A002
        return API.RawKey_d(prefix=prefix, key=key, format=format, convert=convert)



#!/usr/bin/env python3
"""
A Proxy for ChainGuard,
  which allows you to use the default attribute access
  (data.a.b.c)
  even when there might not be an `a.b.c` path in the data.

  Thus:
  data.on_fail(default_value).a.b.c()

  Note: To distinguish between not giving a default value,
  and giving a default value of `None`,
  wrap the default value in a tuple: (None,)
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import atexit#  for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types as types_
import weakref
from copy import deepcopy
from time import sleep
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Proto
from .._base import GuardBase
from .._interface import TomlTypes, ChainProxy_p
from ..errors import GuardedAccessError
from .base import GuardProxy

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
from typing import Never

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(ChainProxy_p)
class GuardFailureProxy(GuardProxy):
    """
    A Wrapper for guarded access to toml values.
    you get the value by calling it.
    Until then, it tracks attribute access,
    and reports that to GuardBase when called.
    It also can type check its value and the value retrieved from the toml data
    """

    def __init__(self, data:GuardBase, types:Any=None, index:Maybe[list[str]]=None, fallback:TomlTypes|Never=Never):
        super().__init__(data, types=types, index=index)
        if fallback == (None,):
            self._fallback = None
        else:
            self._fallback = fallback

        if fallback:
            self._match_type(self._fallback)

    def __call__(self, wrapper:Maybe[callable[[TomlTypes], Any]]=None, fallback_wrapper:Maybe[callable[[TomlTypes], Any]]=None) -> Any:
        """
        Reify a proxy into an actual value, or its fallback.
        Optionally call a wrapper function on the actual value,
        or a fallback_wrapper function on the fallback
        """
        self._notify()
        wrapper : callable[[TomlTypes], TomlTypes] = wrapper or (lambda x: x)
        fallback_wrapper                           = fallback_wrapper or (lambda x: x)
        match self._data, self._fallback:
            case x, y if x is Never and y is Never:
                raise ValueError("No Value, and no fallback")
            case x, None if x is Never or x is None:
                val = None
            case x, data if x is Never or x is None:
                val = fallback_wrapper(data)
            case GuardBase() as data, _:
                val = wrapper(dict(data))
            case _ as data, _:
                val = wrapper(data)

        return self._match_type(val)

    def __getattr__(self, attr:str) -> GuardProxy:
        try:
            match self._data:
                case x if x is Never:
                    raise GuardedAccessError()
                case GuardBase():
                    return self._inject(self._data[attr], attr=attr)
                case _:
                    return self._inject(attr=attr)
        except GuardedAccessError:
            return self._inject(clear=True, attr=attr)

    def __getitem__(self, keys:str|tuple[str]) -> GuardProxy:
        curr = self
        match keys:
            case tuple():
                for key in keys:
                    curr = curr.__getattr__(key)
            case str():
                curr = self.__getattr__(keys)

        return curr

    def _inject(self, val:tuple[Any]=Never, attr:Maybe[str]=None, clear:bool=False) -> GuardProxy:
        match val:
            case _ if clear:
                val = Never
            case x if x is Never:
                val = self._data
            case _:
                pass

        return GuardFailureProxy(val,
                                 types=self._types,
                                 index=self._index(attr),
                                 fallback=self._fallback)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: N805
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import urllib
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
import urllib.parse

if TYPE_CHECKING:
    from jgdv import Maybe, Rx
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv.files.tags import SubstitutionFile

    type UrlParseResult = urllib.parse.ParseResult

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class Bookmark(BaseModel):
    url              : str
    tags             : set[str]              = set()
    name             : str                   = "No Name"
    _tag_sep         : ClassVar[str]         = " : "
    _tag_norm_re     : ClassVar[Rx]          = re.compile(" +")

    @classmethod
    def build[T](cls:type[T], line:str, sep:Maybe[str]=None) -> T:
        """
        Build a bookmark from a line of a bookmark file
        """
        tags : list
        sep  = sep or Bookmark._tag_sep
        tags = []
        match [x.strip() for x in line.split(sep)]:
            case []:
                msg = "Bad line passed to Bookmark"
                raise TypeError(msg)
            case [url]:
                logging.warning("No Tags for: %s", url)
            case [url, *tags]:
                pass

        return Bookmark(url=url,
                        tags=set(tags),
                        sep=sep)

    @field_validator("tags", mode="before")
    def _validate_tags(cls, val:list|set|str) -> set:
        match val:
            case list()|set():
                return { Bookmark._tag_norm_re.sub("_", x.strip()) for x in val }
            case str():
                return { Bookmark._tag_norm_re.sub("_", x.strip()) for x in val.split(Bookmark._tag_sep) }
            case _:
                msg = "Unrecognized tags base"
                raise ValueError(msg, val)

    def __eq__(self, other:object) -> bool:
        match other:
            case Bookmark() as o:
                return self.url == o.url
            case _:
                return False

    def __lt__(self, other:object) -> bool:
        match other:
            case Bookmark() as o:
                return self.url < o.url
            case _:
                return False

    def __str__(self) -> str:
        sep = Bookmark._tag_sep
        tags = sep.join(sorted(self.tags))
        return f"{self.url}{sep}{tags}"

    @property
    def url_comps(self) -> UrlParseResult:
        return urllib.parse.urlparse(self.url)

    def merge(self, other:Bookmark) -> Self:
        """ Merge two bookmarks' tags together,
        creating a new bookmark
        """
        assert(self == other)
        merged = Bookmark(url=self.url,
                          tags=self.tags.union(other.tags),
                          name=self.name)
        return merged

    def clean(self, subs:SubstitutionFile) -> None:
        """
        run tag substitutions on all tags in the bookmark
        """
        cleaned_tags = set()
        for tag in self.tags:
            cleaned_tags.update(subs.sub(tag))

        self.tags = cleaned_tags



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
# ##-- end stdlib imports

from collections import defaultdict
from jgdv import Proto, Mixin
from jgdv._abstract.pre_processable import PreProcessor_p
from . import errors
from . import _interface as API  # noqa: N812

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Callable

if TYPE_CHECKING:
    import enum
    from jgdv import Maybe, MaybeT
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from ._interface import Strang_p
    from jgdv._abstract.pre_processable import PreProcessResult, InstanceData, PostInstanceData
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Vars
HEAD_IDXS : Final[int] = 1
##--| funcs

def name_to_hook(val:str) -> str:
    return f"_{val}_h"

##--| Body

class StrangBasicProcessor[T:Strang_p](PreProcessor_p):
    """ A processor for basic strangs,
    the instance is assigned into Strang._processor

    If the strang type implements _{call}_h,
    the processor uses that for a stage instead
    """
    def use_hook(self, cls:type[T]|T, stage:str, *args:Any, **kwargs:Any) -> MaybeT[bool, tuple]:  # noqa: ANN401
        match cls, getattr(cls, name_to_hook(stage), None):
            case _, None:
                return None
            case _, x if not callable(x):
                return None
            case type(), x:
                assert(callable(x))
                result = x(*args, **kwargs)
            case _, x:
                assert(callable(x))
                result = x(*args, **kwargs)

        match result:
            case None:
                return None
            case bool() as prefer, *rest:
                return (prefer, *rest)
            case x:
                        raise TypeError(type(x))

    def pre_process(self, cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> PreProcessResult[T]:  # noqa: A002, ANN401
        """ run before str.__new__ is called,
        to do early modification of the string
        Filters out extraneous duplicated separators
        """
        text       : str
        inst_data  : InstanceData      = {}
        post_data  : PostInstanceData  = {}
        ctor       : Maybe[type[T]]    = None
        match self.use_hook(cls, "pre_process", input, *args, strict=strict, **kwargs):
            case None:
                case = cls.section(-1).case or ""
                text = case.join(str(x) for x in [input, *args])
            case False, *rest:
                text , inst_data, post_data, ctor = rest  # type: ignore[assignment]
                return text, inst_data, post_data, ctor
            case True, *rest:
                input, inst_data, post_data, ctor = rest  # type: ignore[assignment]  # noqa: A001


        if not self._verify_structure(cls, text):
            raise ValueError(errors.MalformedData, text)

        clean      = self._clean_separators(cls, text).strip()
        val, extracted = self._compress_types(cls, clean)
        post_data.update(extracted)
        match self._get_args(val):
            case int() as args_start:
                post_data['args_start']  = args_start
            case _:
                pass


        return val, inst_data, post_data, None

    def _verify_structure(self, cls:type[T], val:str) -> bool:
        """ Verify basic strang structure.

        ie: all necessary sections are, provisionally, there.
        """
        seps = [x.end for x in cls._sections.order if x.end is not None and x.required]
        return all(x in val for x in seps)

    def _clean_separators(self, cls:type[T], val:str) -> str:
        """ Clean even repetitions of the separator down to single uses

        eg: for sep='.',
        a..b::c....d -> a.b::c.d
        but:
        a.b::c...d -> a.b::c..d
        """
        # TODO join the seps
        seps = [x.case for x in cls._sections.order]
        sep = seps[0] or ""
        sep_double = re.escape(sep * 2)
        clean_re   = re.compile(f"{sep_double}+")
        # Don't reuse sep_double, as thats been escaped
        cleaned    = clean_re.sub(sep * 2, val)
        trimmed    = cleaned.removesuffix(sep).removesuffix(sep)
        return trimmed

    def _compress_types(self, cls:type[T], val:str) -> tuple[str, dict]:  # noqa: ARG002
        """ Extract values of explicitly typed words.

        allows the base str of the Strang to be readable,
        and for post-process to insert types as necessary

        eg: a.b.c::d.e.<uuid:....> -> (a.b.c::d.e.<uuid>, {uuids:[UUIDstr]}

        """
        curr       : re.Match
        text       : list                          = []
        extracted  : list[tuple[str, Maybe[str]]]  = []
        idx        : int                           = 0
        for curr in API.TYPE_ITER_RE.finditer(val):
            match curr.groups():
                case ["<", str() as key, str() as oval, ">"]:
                    extracted.append((key, oval))
                    _,start         = curr.span(2)
                    rest,end        = curr.span(4)
                    text.append(val[idx:start])
                    text.append(val[rest:end])
                    idx = end
                case ["<", str() as key, None, ">"]:
                    extracted.append((key, None))
        else:
            text.append(val[idx:])
            return "".join(text), {'types': extracted}

    def _get_args(self, val:str) -> Maybe[int]:
        try:
            idx : int = val.rindex(API.ARGS_CHARS[0])
            assert(val[-1] == API.ARGS_CHARS[-1])
            assert(API.ARGS_RE.match(val[idx:]))
        except ValueError:
            return None
        else:
            return idx

    ##--|

    def process(self, obj:T, *, data:PostInstanceData) -> Maybe[T]:
        """ slice the sections of the strang

        populates obj.data:
        - slices
        - flat
        - bounds
        """
        pos_offset    : int
        word_indices  : list[tuple[int, ...]]
        sec_slices    : list[slice]
        flat_slices   : list[slice]
        match self.use_hook(obj, "process", data=data):
            case None:
                pass
            case True, x:
                assert(isinstance(x, type(obj)|None))
                return x
            case False, None:
                pass
            case False, x:
                assert(isinstance(x, type(obj)|None))
                obj = x

        logging.debug("Processing Strang: %s", str.__str__(obj))
        match data:
            case {"args_start": int() as arg_s}:
                obj.data.args_start = arg_s
            case _:
                pass

        pos_offset, index_offset = 0, 0
        sec_slices, flat_slices, word_indices = [], [], []
        for section in obj.sections():
            sec, words, extend = self._process_section(obj, section, start=pos_offset)
            sec_slices.append(sec)
            word_indices.append(tuple(range(index_offset, index_offset+len(words))))
            index_offset += len(words)
            flat_slices  += words
            pos_offset    = sec.stop + extend
        else:
            obj.data.sec_words  = tuple(word_indices)
            obj.data.flat_idx   = tuple((i,j) for i,x in enumerate(obj.data.sec_words) for j in range(len(x)))
            obj.data.sections   = tuple(sec_slices)
            obj.data.words      = tuple(flat_slices)
            self._process_args(obj, data=data)
            return None

    def _process_section(self, obj:T, section:API.Sec_d, *, start:int=-1) -> tuple[slice, tuple[slice, ...], int]:
        """ Set the slices of a section, return the index where the section ends """
        word_slices   : tuple[slice]
        search_end    : int  = obj.data.args_start or len(obj)
        bound_extend  : int  = 0
        match section.end:
            case str() as x:
                try:
                    bound_extend = len(x)
                    search_end   = obj.index(x, start=start)
                except (ValueError, TypeError):
                    return slice(start, start), (), 0
            case None:
                pass
        ##--|

        word_slices = self._slice_section(obj,
                                          case=[section.case, section.end],
                                          start=start,
                                          max=search_end)
        assert(all((start <= x.start <= x.stop <= search_end) for x in word_slices))
        match word_slices:
            case []:
                return slice(start, search_end), (), 0
            case _:
                return slice(start, search_end), word_slices, bound_extend

    def _slice_section(self, obj:T, *, case:list[Maybe[str]], start:int=0, max:int=-1) -> tuple[slice]:  # noqa: A002
        """ Get a list of word slices of a section, with an offset. """
        curr    : re.Match
        slices  : list[slice]  = []
        end                    = max or len(obj)
        escaped                = "|".join(re.escape(x) for x in case if x is not None)
        reg                    = re.compile(f"(.*?)({escaped}|$)")
        words                  = []
        for curr in reg.finditer(cast("str", obj), start, end):
            span = curr.span(1)
            if span[0] == end:
                continue
            slices.append(slice(*span))
            words.append(obj[span[0]:span[1]])
        else:
            return cast("tuple[slice]", tuple(slices))

    def _process_args(self, obj:T, *, data:dict) -> None:
        """ Extract args and set values as necessary """
        if not (arg_s:=obj.data.args_start):
            return

        selection = sorted([x.strip() for x in API.STRGET(obj, slice(arg_s+1, -1)).split(API.ARGS_CHARS[1])])
        if len(selection) != len(set(selection)):
            raise ValueError(selection)

        obj.data.args = tuple(selection)
        if API.UUID_WORD in selection:
            assert('types' in data), data
            match data['types'].pop():
                case "uuid", str() as uid_val:
                    obj.data.uuid = UUID(uid_val)
                case "uuid", None:
                    obj.data.uuid = uuid1()
                case _:
                    pass

    ##--|

    def post_process(self, obj:T, data:PostInstanceData) -> Maybe[T]:
        """ With the strang cleaned and slices, build meta data for words

        takes the data extracted during pre-processing.

        """
        metas  : list  = []
        if 'types' in data:
            data['types'].reverse()

        match self.use_hook(obj, "process", data=data):
            case None:
                pass
            case True, x:
                assert(isinstance(x, type(obj)|None))
                return x
            case False, None:
                pass
            case False, x:
                assert(isinstance(x, type(obj)|None))
                obj = x

        logging.debug("Post-processing Strang: %s", str.__str__(obj))
        for i in range(len(obj.sections())):
            metas += self._post_process_section(obj, i, data)
        else:
            obj.data.meta = tuple(metas)  # type: ignore[assignment]
            self._validate_marks(obj)
            self._calc_obj_meta(obj)
            return None

    def _post_process_section(self, obj:T, idx:int, data:dict) -> list:
        type MetaTypes              = Maybe[UUID|API.StrangMarkAbstract_e|int]
        elem     : str
        section  : API.Sec_d        = obj.section(idx)
        count    : int              = len(obj.data.sec_words[idx])
        meta     : list[MetaTypes]  = [None for x in range(count)]
        for i, word_idx in enumerate(obj.data.sec_words[idx]):
            elem                    = obj[obj.data.words[word_idx]]
            assert(isinstance(elem, str))
            # Discriminate the str
            match elem:
                case x if (mark_elem:=self._implicit_mark(x, sec=section, data=data, index=i, maxcount=count)) is not None:
                    logging.debug("(%s) Found Named Marker: %s", i, mark_elem)
                    meta[i] = mark_elem
                case x if (type_mark:=self._make_type(x, sec=section, data=data, obj=obj)) is not None:
                    meta[i] = type_mark
                case x if (mark_elem:=self._build_mark(x, sec=section, data=data)) is not None:
                    logging.debug("(%s) Found Named Marker: %s", i, mark_elem)
                    meta[i] = mark_elem
                case _: # nothing special
                    pass
        else:
            return meta

    def _calc_obj_meta(self, obj:T) -> None:
        """ Set object level meta dict

        ie: mark the obj as an instance
        """
        pass

    def _validate_marks(self, obj:T) -> None:
        """ Check marks make sense.
        eg: +|_ are only at obj[1:0]

        """
        pass

    ##--| utils

    def _make_type(self, val:str, *, sec:API.Sec_d, data:dict, obj:T) -> Maybe[Any]:  # noqa: ARG002
        """ Handle <type> words, which may have had data extracted during pre-processing.

        """
        result : Maybe    = None
        if not (word:=API.TYPE_RE.match(val)):
            return None

        match data.get('types', [None]).pop():
            case None: # No types data remains
                raise ValueError()
            case str() as key, typeval:
                pass

        match word.groups()[0], typeval:
            case x, _ if x != key: # Mismatch between types
                raise ValueError(x, key)
            case "uuid", None:
                result = uuid1()
            case "uuid", str() as spec:
                result = UUID(spec)
            case "int", str() as spec:
                result = int(spec)
            case [x, _]:
                raise ValueError()

        ##--|
        return result

    def _build_mark(self, val:str, *, sec:API.Sec_d, data:dict) -> Maybe[API.StrangMarkAbstract_e]:  # noqa: ARG002
        """ converts applicable words to mark enum values
        Matches using strang._interface.MARK_RE

        """
        match sec.marks:
            case None:
                return None
            case x:
                marks = x
        match API.MARK_RE.match(val):
            case re.Match() as matched if (key:=matched[1]) is not None:
                if key.lower() in marks:
                    return marks(key)
                return None
            case _:
                return None

    def _implicit_mark(self, val:str, *, sec:API.Sec_d, data:dict, index:int, maxcount:int) -> Maybe[API.StrangMarkAbstract_e]:  # noqa: ARG002
        """ Builds certain implicit marks,
        but only for the first and last words of a section

        # TODO handle combined marks like val::+_.blah

        """
        first_or_last = index in {0, maxcount-1}
        match sec.marks:
            case None:
                return None
            case x:
                marks = x
        match marks.skip():
            case None:
                pass
            case x if val == x:
                return x

        if not (first_or_last and val in marks):
            return None
        return marks(val)




#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

from jgdv import JGDVError

class ChainGuardError(JGDVError):
    pass

class GuardedAccessError(AttributeError, ChainGuardError):
    pass



"""
Strang, a Structured String class.

Strangs are str's, and can be used as str's.
But they validate their format, and allow access to sub parts. eg::

    v = Strang("group::body.name")
    v[0:] == "group"
    v[1:] == "body.name"
    v[1:0] == "body"
    v[1:1] == "name"

There is also a specialized strang CodeReference.
for easily importing code::

    v = CodeReference("fn::jgdv.identity_fn")
    assert(callable(v()))
    assert(v() == jgdv.identity_fn)

"""

from ._interface import Strang_p, Importable_p
from .strang import Strang
from .code_ref import CodeReference
from .errors import StrangError



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports


# ##-- 1st party imports
from jgdv.files.bookmarks.bookmark import Bookmark

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class BookmarkCollection(BaseModel):

    entries : list[Bookmark] = []
    ext     : str            = ".bookmarks"

    @staticmethod
    def read(fpath:pl.Path) -> BookmarkCollection:
        """ Read a file to build a bookmark collection """
        bookmarks = BookmarkCollection()
        for line in (x.strip() for x in fpath.read_text().split("\n")):
            if not bool(line):
                continue
            bookmarks += Bookmark.build(line)

        return bookmarks

    def __str__(self):
        return "\n".join(map(str, sorted(self.entries)))

    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __iadd__(self, value):
        return self.update(value)

    def __iter__(self):
        return iter(self.entries)

    def __contains__(self, value:Bookmark) -> bool:
        return value in self.entries

    def __len__(self):
        return len(self.entries)

    def __hash__(self):
        return id(self)

    def update(self, *values) -> Self:
        for val in values:
            match val:
                case Bookmark():
                    self.entries.append(val)
                case BookmarkCollection():
                    self.entries += val.entries
                case [*vals] | set(vals):
                    self.update(*vals)
                case _:
                    raise TypeError(type(val))
        return self

    def difference(self, other:Self) -> Self:
        result = BookmarkCollection()
        for bkmk in other:
            if bkmk not in self:
                result += bkmk

        return result

    def merge_duplicates(self):
        deduplicated = {}
        for x in self:
            if x.url not in deduplicated:
                deduplicated[x.url] = x
            else:
                deduplicated[x.url] = x.merge(deduplicated[x.url])

        self.entries = list(deduplicated.values())



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, ANN001, B011, PLR0133, PLR2004
# Imports:
from __future__ import annotations

import itertools as itz
import logging as logmod
import warnings
import pathlib as pl
import pytest
import types
from typing import Any
from ... import ParseError
from .. import param_spec as pbase
from ..param_spec import ParamSpec, ParamProcessor
from ... import _interface as API # noqa: N812

##--| vars
logging          = logmod.root
good_names       = ("test", "blah", "bloo")
parse_test_vals  = [("-test", "-", "test"),
                    ("--blah", "--", "blah"),
                    ("--bloo=", "--", "bloo"),
                    ("+aweg", "+", "aweg"),
                   ]

sorting_names   = ["-next", "<>another", "--test", "<2>other", "<1>diff"]
correct_sorting = ["-next", "--test", "diff", "other", "another"]
##--|

class TestParamProcessor:

    def test_sanity(self):
        assert(True is not False)

    ##--| name parsing

    @pytest.mark.parametrize(["full", "pre", "name"], parse_test_vals)
    def test_name_parse(self, full, pre, name):
        res = ParamProcessor().parse_name(full)
        assert(res['name'] == name)
        assert(res['prefix'] == pre)

    def test_name_parse_complex(self):
        res = ParamProcessor().parse_name("--group-by")
        assert(res['name'] == "group-by")
        assert(res['prefix'] == "--")

    def test_name_parse_complex_assign(self):
        res = ParamProcessor().parse_name("--group-by=")
        assert(res['name'] == "group-by")
        assert(res['prefix'] == "--")

    ##--| match utils

    @pytest.mark.parametrize("key", [*good_names])
    def test_match_on_head(self, key):
        """ Simple names match a set of standard variations """
        obj    = ParamProcessor()
        param  = ParamSpec(name=f"-{key}")
        assert(obj.matches_head(param, f"-{key}"))
        assert(obj.matches_head(param, f"-{key[0]}"))
        assert(obj.matches_head(param, f"-no-{key}"))

    @pytest.mark.parametrize("key", [*good_names])
    def test_match_on_head_assignments(self, key):
        """ Assignments match on a reduced set of variations """
        obj    = ParamProcessor()
        param  = ParamSpec(name=f"--{key}=")
        assert(obj.matches_head(param, f"--{key}=val"))
        assert(obj.matches_head(param, f"--{key[0]}=val"))

    @pytest.mark.parametrize("key", [*good_names])
    def test_match_on_head_fail(self, key):
        """ some variations do not match """
        obj    = ParamProcessor()
        param  = ParamSpec(name=f"--{key}")
        assert(not obj.matches_head(param, key))
        assert(not obj.matches_head(param, f"{key}=blah"))
        assert(not obj.matches_head(param, f"-{key}=val"))
        assert(not obj.matches_head(param, f"-{key[0]}=val"))

##--|

class TestParamSpec_classmethods:

    def test_sanity(self):
        assert(True is not False)

    def test_build_defaults(self):
        param_dicts = [
            {"name":"test","default":"test"},
            {"name":"-next", "default":2},
            {"name":"--other", "default":list},
            {"name":"+another", "default":lambda: [1,2,3,4]},
        ]
        params = [y for x in param_dicts if (y:=ParamSpec(**x))]
        assert(len(params) == len(param_dicts))
        result = ParamSpec.build_defaults(params)
        assert(result['test'] == 'test')
        assert(result['next'] == 2)
        assert(result['other'] == [])
        assert(result['another'] == [1,2,3,4])

    def test_check_insist_params(self):
        param_dicts = [
            {"name":"test","default":"test", "insist":False},
            {"name":"next", "default":2, "insist":True},
            {"name":"other", "default":list, "insist":True},
            {"name":"another", "default":lambda: [1,2,3,4], "insist":False},
        ]
        params = [ParamSpec(**x) for x in param_dicts]
        ParamSpec.check_insists(params, {"next": 2, "other":[1,2,3]})
        assert(True)

    def test_check_insist_params_fail(self):
        param_dicts = [
            {"name":"test","default":"test", "insist":False},
            {"name":"next", "default":2, "insist":True},
            {"name":"other", "default":list, "insist":True},
            {"name":"another", "default":lambda: [1,2,3,4], "insist":False},
        ]
        params = [ParamSpec(**x) for x in param_dicts]
        with pytest.raises(ParseError) as ctx:
            ParamSpec.check_insists(params, {"other":[1,2,3]})

        assert(ctx.value.args[-1] == ["next"])

class TestParamSpec_Sorting:

    def test_sanity(self):
        assert(True is not False)

    def test_sorting(self):
        target_sort  = correct_sorting
        param_dicts  = [{"name":x} for x in sorting_names]
        params       = [ParamSpec(**x) for x in param_dicts]
        s_params     = sorted(params, key=ParamSpec.key_func)
        for x,y in zip(s_params, target_sort, strict=True):
            assert(x.key_str == y), s_params

class TestParamSpec_Basic:

    def test_sanity(self):
        assert(True is not False)

    def test_initial(self):
        match ParamSpec(name="test"):
            case ParamSpec() as obj:
                assert(isinstance(obj, API.ParamStruct_p))
            case x:
                 assert(False), x


    def test_equality(self):
        obj1 = ParamSpec(name="test")
        obj2 = ParamSpec(name="test")
        assert(obj1 is not obj2)
        assert(obj1 == obj2)


    def test_equality_trivial_fail(self):
        obj1 = ParamSpec(name="test")
        obj2 = "blah"
        assert(obj1 is not obj2)
        assert(obj1 != obj2)


    def test_equality_fail_on_name(self):
        obj1 = ParamSpec(name="test")
        obj2 = ParamSpec(name="blah")
        assert(obj1 is not obj2)
        assert(obj1 != obj2)


    def test_equality_fail_on_prefix(self):
        obj1 = ParamSpec(name="-test")
        obj2 = ParamSpec(name="--test")
        assert(obj1 is not obj2)
        assert(obj1 != obj2)


    def test_equaltiy_fail_on_separator(self):
        obj1 = ParamSpec(name="--test=")
        obj2 = ParamSpec(name="--test")
        assert(obj1 is not obj2)
        assert(obj1 != obj2)

    def test_key_strs_prop(self):
        obj = ParamSpec(name="-test")
        assert(obj.key_str == "-test")
        match obj.key_strs:
            case list():
                assert(True)
            case x:
                 assert(False), x

    @pytest.mark.parametrize(["key", "prefix"], zip(good_names, itz.cycle(["-", "--"])))
    def test_short_key(self, key, prefix):
        obj = ParamSpec(name=f"{prefix}{key}")
        assert(obj.short == key[0])
        match prefix:
            case "--":
                assert(obj.short_key_str == f"{prefix}{key[0]}")
            case "-":
                assert(obj.short_key_str == f"{prefix}{key[0]}")

class TestParamSpec_Types:

    def test_sanity(self):
        assert(True is not False)

    def test_int(self):
        obj = ParamSpec(**{"name":"blah", "type":int})
        assert(obj.type_ is int)
        assert(obj.default == 0)

    def test_Any(self): # noqa: N802
        obj = ParamSpec(**{"name":"blah", "type":Any})
        assert(obj.type_ is None)
        assert(obj.default is None)

    def test_typed_list(self):
        obj = ParamSpec(**{"name":"blah", "type":list[str]})
        assert(isinstance(obj.type_, type))
        assert(obj.type_ is list)
        assert(obj.default is list)

    def test_annotated(self):

        class TestParam(ParamSpec[str]):
            pass

        assert(issubclass(TestParam, ParamSpec))
        obj = TestParam(name="blah")
        assert(obj.type_ is str)
        assert(obj.default == '')

    def test_annotated_list(self):

        class TestParam2(ParamSpec[list[str]]):
            pass

        obj = TestParam2(name="blah")
        assert(obj.type_ is list)
        assert(obj.default is list)

    def test_type_fail(self):
        with pytest.raises(TypeError):
            ParamSpec(name="blah", type=ParamSpec)

    def test_type_build_fail(self):
        with pytest.raises(TypeError):
            ParamSpec(**{"name":"blah", "type":ParamSpec})

class TestParamSpec_Building:

    def test_sanity(self):
        assert(True is not False)

    def test_int(self):
        obj = ParamSpec(**{"name":"blah", "type":int})
        assert(obj.type_ is int)
        assert(obj.default == 0)

    def test_Any(self): # noqa: N802
        obj = ParamSpec(**{"name":"blah", "type":Any})
        assert(obj.type_ is None)
        assert(obj.default is None)

    def test_typed_list(self):
        obj = ParamSpec(**{"name":"blah", "type":list[str]})
        assert(isinstance(obj.type_, type))
        assert(obj.type_ is list)
        assert(obj.default is list)

    def test_type_fail(self):
        with pytest.raises(TypeError):
            ParamSpec(name="-blah", type=ParamSpec)

    def test_type_build_fail(self):
        with pytest.raises(TypeError):
            ParamSpec(**{"name":"-blah", "type":ParamSpec})

class TestParamSpec_Consumption:

    def test_sanity(self):
        assert(True is not False)

    def test_consume_nothing(self):
        data     = {"name" : "test"}
        in_args  = []
        obj      = ParamSpec(**data)
        match obj.consume(in_args):
            case None:
                assert(True)
            case _:
                assert(False)

    def test_consume_match(self):
        data     = {"name" : "-test"}
        in_args  = ["-test"]
        obj      = ParamSpec(**data)
        assert(isinstance(obj, ParamSpec))
        match obj.consume(in_args):
            case {"test": True}, 1:
                assert(True)
            case _:
                assert(False)

    def test_consume_with_remaining(self):
        data     = {"name"  : "-test"}
        in_args  = ["-test", "blah"]
        obj      = ParamSpec(**data)
        assert(isinstance(obj, ParamSpec))
        match obj.consume(in_args):
            case {"test": True}, 1:
                assert(True)
            case _:
                assert(False)

    def test_consume_with_offset(self):
        data     = {"name" : "-test"}
        in_args  = ["blah", "bloo", "-test", "aweg"]
        obj      = ParamSpec(**data)
        match obj.consume(in_args, offset=2):
            case {"test": True}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_nothing_without_offset(self):
        data     = {"name" : "-test", "type" : str}
        in_args  = ["blah", "bloo", "-test", "aweg"]
        obj      = ParamSpec(**data)
        assert(obj.type_ is str)
        match obj.consume(in_args):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_consume_short(self):
        data     = {"name" : "--test", "default":False, "type":bool}
        in_args  = ["--t", "blah", "bloo"]
        obj      = ParamSpec(**data)
        assert(obj.type_ is bool)
        match obj.consume(in_args):
            case {"test": True}, 1:
                assert(True)
            case x:
                assert(False), x




#!/usr/bin/env python3
"""

"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

import stackprinter

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe, RxStr
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

# Global Vars:

# Body:

class StackFormatter_m:
    """ A Mixin Error formatter, adapted from stackprinter's docs
    Compactly Formats the error stack trace, without src.

    """

    indent_str       : ClassVar[str]         = "  |  "
    suppress         : ClassVar[list[RxStr]] = [r".*pydantic.*", r"<frozen importlib._bootstrap>"]
    source_height    : ClassVar[int]         = 10
    source_lines     : ClassVar[int|str]     = 0
    use_stackprinter : bool                  = True

    def formatException(self, exc_info):
        match exc_info:
            case None | (None, None, None):
                return ""
            case _ if not self.use_stackprinter:
                return super().formatException(exc_info)

        msg : str = stackprinter.format(exc_info,
                                        source_lines=self.source_lines,
                                        suppressed_paths=self.suppress)
        lines = [x for x in msg.splitlines() if bool(x)]
        indented = [f"{self.indent_str}{line}\n" for line in lines[-self.source_height:]]
        return "".join(indented)

    def formatStack(self, stack_info):
        return super().formatStack(stack_info)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

##--|
from ..param_spec import ParamSpec
from .. import extra
##--|

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

@pytest.mark.skip
class TestRepeatableParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_consume_list_single_value(self):
        obj = extra.RepeatableParam(**{"name" : "test", "type" : list})
        match obj.consume(["-test", "bloo"]):
            case {"test": ["bloo"]}, 2:
                assert(True)
            case x:
                assert(False), x

    def test_consume_list_multi_key_val(self):
        obj     = extra.RepeatableParam(**{"name":"test"})
        in_args = ["-test", "bloo", "-test", "blah", "-test", "bloo", "-not", "this"]
        match obj.consume(in_args):
            case {"test": ["bloo", "blah", "bloo"]}, 6:
                assert(True)
            case x:
                assert(False), x

    def test_consume_set_multi(self):
        obj = extra.RepeatableParam[set](**{
            "name"    : "test",
            "type"    : set,
            "default" : set,
          })
        in_args             = ["-test", "bloo", "-test", "blah", "-test", "bloo", "-not", "this"]
        match obj.consume(in_args):
            case {"test": set() as x}, 6:
                assert(x == {"bloo", "blah"})
            case x:
                assert(False), x

    def test_consume_str_multi_set_fail(self):
        obj = extra.RepeatableParam[set](**{
            "name" : "test",
            "type" : str,
            "default" : "",
          })
        in_args             = ["-nottest", "bloo", "-nottest", "blah", "-nottest", "bloo", "-not", "this"]
        match obj.consume(in_args):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_consume_multi_assignment_fail(self):
        obj     = extra.RepeatableParam(**{"name":"test", "type":list, "default":list, "prefix":"--"})
        in_args = ["--test=blah", "--test=bloo"]
        match obj.consume(in_args):
            case None:
                assert(True)
            case x:
                assert(False), x

@pytest.mark.skip
class TestImplicitParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

@pytest.mark.skip
class TestChoiceParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

@pytest.mark.skip
class TestEntryParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

@pytest.mark.skip
class TestConstrainedParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

@pytest.mark.skip
class TestWildCardParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_wildcard_assign(self):
        obj = extra.WildcardParam()
        match obj.consume(["--blah=other"]):
            case {"blah":"other"}, 1:
                assert(True)
            case x:
                assert(False), x
